# -*- coding: utf-8 -*-
"""
Main proposal object that includes normalising flows.
"""
import copy
import datetime
import logging
import os

import matplotlib.pyplot as plt
import numpy as np
import numpy.lib.recfunctions as rfn
import torch

from ..flowmodel import FlowModel, update_config
from ..livepoint import (
    live_points_to_array,
    numpy_array_to_live_points,
    get_dtype,
    DEFAULT_FLOAT_DTYPE
    )
from ..reparameterisations import (
    CombinedReparameterisation,
    NullReparameterisation,
    get_reparameterisation
    )
from ..plot import plot_live_points, plot_1d_comparison
from .rejection import RejectionProposal
from ..utils import (
    get_uniform_distribution,
    get_multivariate_normal,
    detect_edge,
    configure_edge_detection,
    save_live_points
    )

logger = logging.getLogger(__name__)


class FlowProposal(RejectionProposal):
    """
    Object that handles training and proposal points

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User defined model.
    flow_config : dict, optional
        Configuration for training the normalising flow. If None, uses default
        settings. Defaults to None.
    output : str, optional
        Path to output directory.
    plot : {True, False, 'all', 'min'}, optional
        Controls the plotting level: ``True`` - all plots; ``False`` - no
        plots; ``'all'`` -  all plots and ``'min'`` -  1d plots and loss.
    latent_prior : {'truncated_gaussian', 'gaussian', 'uniform_nsphere', \
            'gaussian'}, optional
        Prior distribution in the latent space. Defaults to
        'truncated_gaussian'.
    poolsize : int, optional
        Size of the proposal pool. Defaults to 10000.
    update_poolsize : bool, optional
        If True the poolsize is updated using the current acceptance of the
        nested sampler.
    max_poolsize_scale : int, optional
        Maximum scale for increasing the poolsize. E.g. if this value is 10
        and the poolsize is 1000 the maximum number of points in the pool
        is 10,000.
    drawsize : int, optional
        Number of points to simultaneosly draw when populating the proposal
        Defaults to 10000
    check_acceptance : bool, optional
        If True the acceptance is computed after populating the pool. This
        includes computing the likelihood for every point. Default False.
    min_radius : float, optional
        Minimum radius used for popluation. If not specified not minimum is
        used.
    max_radius : float, optional
        If a float then this value is used as an upper limit for the
        computed radius when populating the proposal. If unspecified no
        upper limit is used.
    fixed_radius : float, optional
        If specified and the chosen latent distribution is compatible, this
        radius will be used to draw new samples instead of the value computed
        with the flow.
    compute_radius_with_all : bool, optional
        If True all the radius of the latent contour is computed using the
        maximum radius of all the samples used to train the flow.
    fuzz : float, optional
        Fuzz-factor applied to the radius. If unspecified no fuzz-factor is
        applied.
    expansion_fraction : float, optional
        Similar to ``fuzz`` but instead a scaling factor applied to the radius
        this specifies a rescaling for volume of the n-ball used to draw
        samples. This is translated to a value for ``fuzz``.
    zero_reset : int, optional
        Value used when reseting proposal if zero samples are accepted.
        If specified is after drawing samples zero_reset times the current
        poolsize is less than half, then the radius is reduced by 1% and
        the pool is reset.
    truncate : bool, optional
        Truncate proposals using probability compute for worst point.
        Not recommended.
    rescale_parameters : list or bool, optional
        If True live points are rescaled to `rescale_bounds` before training.
        If an instance of `list` then must contain names of parameters to
        rescale. If False no rescaling is applied.
    rescale_bounds : list, optional
        Lower and upper bound to use for rescaling. Defaults to [-1, 1]. See
        `rescale_parameters`.
    update_bounds : bool, optional
        If True bounds used for rescaling are updated at the starting of
        training. If False prior bounds are used.
    boundary_inversion : bool or list, optional
        If True boundary inversion is applied to all bounds. If
        If an instance of `list` of parameters names, then inversion
        only applied to these parameters. If False (default )no inversion is
        used.
    inversion_type : {'split', 'duplicate'}
        Type of inversion to use. ``'split'`` keeps the number of samples
        the sample but mirrors half around the bound. ``'duplicate'`` mirrors
        all the samples at the bound.
    detect_edges : bool, optional
        If True, when applying the version the option of no inversion is
        allowed.
    detect_edges_kwargs : dict, optional
        Dictionary of keyword arguments passed to \
                :func:`nessai.utils.detect_edge`.
    reparameterisations : dict, optional
        Dictionary for configure more flexible reparameterisations. This
        ignores any of the other settings related to rescaling. For more
        details see the documentation.
    keep_samples : bool, optional
        If true samples are stored when repopulating the propolal. Not
        recommended.
    n_pool : int, optional
        Number of threads to use for evaluating the likelihood.
    draw_latent_kwargs : dict, optional
        Dictionary of kwargs passed to the function for drawing samples
        in the latent space. See the functions in utils for the possible
        kwargs.
    """

    def __init__(self,
                 model,
                 flow_config=None,
                 output='./',
                 poolsize=None,
                 rescale_parameters=True,
                 latent_prior='truncated_gaussian',
                 fuzz=1.0,
                 keep_samples=False,
                 plot='min',
                 fixed_radius=False,
                 drawsize=None,
                 check_acceptance=False,
                 truncate=False,
                 zero_reset=None,
                 rescale_bounds=[-1, 1],
                 expansion_fraction=1.0,
                 boundary_inversion=False,
                 inversion_type='split',
                 update_bounds=True,
                 min_radius=False,
                 max_radius=50.0,
                 pool=None,
                 n_pool=None,
                 max_poolsize_scale=10,
                 update_poolsize=True,
                 save_training_data=False,
                 compute_radius_with_all=False,
                 draw_latent_kwargs=None,
                 detect_edges=False,
                 detect_edges_kwargs=None,
                 reparameterisations=None,
                 **kwargs):

        super(FlowProposal, self).__init__(model)
        logger.debug('Initialising FlowProposal')

        self._x_dtype = False
        self._x_prime_dtype = False

        self.flow = None
        self._flow_config = None
        self.populated = False
        self.populating = False    # Flag used for resuming during population
        self.indices = []
        self.training_count = 0
        self.populated_count = 0
        self.names = []
        self.training_data = None
        self.save_training_data = save_training_data
        self.x = None
        self.samples = None
        self.rescaled_names = []
        self.acceptance = []
        self.approx_acceptance = []
        self._edges = {}
        self._reparameterisation = None
        self.use_x_prime_prior = False

        self.reparameterisations = reparameterisations

        self.output = output

        self.configure_population(poolsize, drawsize, update_poolsize,
                                  max_poolsize_scale, fuzz, expansion_fraction,
                                  latent_prior)

        self.rescale_parameters = rescale_parameters
        self.keep_samples = keep_samples
        self.update_bounds = update_bounds
        self.check_acceptance = check_acceptance
        self.rescale_bounds = rescale_bounds
        self.truncate = truncate
        self.zero_reset = zero_reset
        self.boundary_inversion = boundary_inversion
        self.inversion_type = inversion_type
        self.flow_config = flow_config

        self.detect_edges = detect_edges
        self.detect_edges_kwargs = \
            configure_edge_detection(detect_edges_kwargs, detect_edges)

        self.compute_radius_with_all = compute_radius_with_all
        self.configure_fixed_radius(fixed_radius)
        self.configure_min_max_radius(min_radius, max_radius)

        self.pool = pool
        self.n_pool = n_pool

        self.configure_plotting(plot)

        self.clip = self.flow_config.get('clip', False)

        if draw_latent_kwargs is None:
            self.draw_latent_kwargs = {}
        else:
            self.draw_latent_kwargs = draw_latent_kwargs
        self.configure_latent_prior()
        self.alt_dist = None

        if kwargs:
            kwargs.pop('max_threads', None)
            logger.warning(
                f'Extra kwargs were parsed to FlowProposal: {kwargs}')

    @property
    def poolsize(self):
        """
        Return the poolsize based of the base value and the current
        value of the scaling
        """
        return int(self._poolsize_scale * self._poolsize)

    @property
    def flow_config(self):
        """Return the configuration for the flow"""
        return self._flow_config

    @flow_config.setter
    def flow_config(self, config):
        """Set configuration (includes checking defaults)"""
        self._flow_config = update_config(config)

    @property
    def dims(self):
        """Return the number of dimensions"""
        return len(self.names)

    @property
    def rescaled_dims(self):
        """Return the number of rescaled dimensions"""
        return len(self.rescaled_names)

    @property
    def x_dtype(self):
        """Return the dtype for the x space"""
        if not self._x_dtype:
            self._x_dtype = get_dtype(self.names, DEFAULT_FLOAT_DTYPE)
        return self._x_dtype

    @property
    def x_prime_dtype(self):
        """Return the dtype for the x prime space"""
        if not self._x_prime_dtype:
            self._x_prime_dtype = \
                get_dtype(self.rescaled_names, DEFAULT_FLOAT_DTYPE)
        return self._x_prime_dtype

    @property
    def population_dtype(self):
        """
        dtype used for populating the proposal, depends on if the prior
        is defined in the x space or x-prime space
        """
        if self.use_x_prime_prior:
            return self.x_prime_dtype
        else:
            return self.x_dtype

    def configure_population(self, poolsize, drawsize, update_poolsize,
                             max_poolsize_scale, fuzz, expansion_fraction,
                             latent_prior):
        """
        Configure settings related to population
        """
        if poolsize is None:
            raise RuntimeError('Must specify a poolsize!')

        if drawsize is None:
            drawsize = poolsize

        self._poolsize = poolsize
        self._poolsize_scale = 1.0
        self.update_poolsize = update_poolsize
        self.max_poolsize_scale = max_poolsize_scale
        self.ns_acceptance = 1.
        self.drawsize = drawsize
        self.fuzz = fuzz
        self.expansion_fraction = expansion_fraction
        self.latent_prior = latent_prior

    def configure_plotting(self, plot):
        """Configure plotting"""
        if plot:
            if isinstance(plot, str):
                if plot == 'all':
                    self._plot_pool = 'all'
                    self._plot_training = 'all'
                elif plot == 'train':
                    self._plot_pool = False
                    self._plot_training = 'all'
                elif plot == 'pool':
                    self._plot_pool = True
                    self._plot_training = False
                elif plot == 'minimal' or plot == 'min':
                    self._plot_pool = True
                    self._plot_training = True
                else:
                    logger.warning(
                        f'Unknown plot argument: {plot}, setting all false'
                        )
                    self._plot_pool = False
                    self._plot_training = False
            else:
                self._plot_pool = True
                self._plot_training = True

        else:
            self._plot_pool = False
            self._plot_training = False

    def configure_latent_prior(self):
        """Configure the latent prior"""
        if self.latent_prior == 'truncated_gaussian':
            from ..utils import draw_truncated_gaussian
            self.draw_latent_prior = draw_truncated_gaussian
            var = self.flow_config['model_config'].get('kwargs', {}).get('var')
            if var and 'var' not in self.draw_latent_kwargs:
                self.draw_latent_kwargs['var'] = var

        elif self.latent_prior == 'gaussian':
            logger.warning('Using a gaussian latent prior WITHOUT truncation')
            from ..utils import draw_gaussian
            self.draw_latent_prior = draw_gaussian
        elif self.latent_prior == 'uniform':
            from ..utils import draw_uniform
            self.draw_latent_prior = draw_uniform
        elif self.latent_prior in ['uniform_nsphere', 'uniform_nball']:
            from ..utils import draw_nsphere
            self.draw_latent_prior = draw_nsphere
        else:
            raise RuntimeError(
                f'Unknown latent prior: {self.latent_prior}, choose from: '
                'truncated_gaussian (default), gaussian, '
                'uniform, uniform_nsphere'
                )

    def configure_fixed_radius(self, fixed_radius):
        """Configure the fixed radius"""
        if fixed_radius:
            try:
                self.fixed_radius = float(fixed_radius)
            except ValueError:
                logger.error(
                    'Fixed radius enabled but could not be converted to a '
                    'float. Setting fixed_radius=False'
                    )
                self.fixed_radius = False
        else:
            self.fixed_radius = False

    def configure_min_max_radius(self, min_radius, max_radius):
        """
        Configure the mininum and maximum radius
        """
        if isinstance(min_radius, (int, float)):
            self.min_radius = float(min_radius)
        else:
            raise RuntimeError('Min radius must be an int or float')

        if max_radius:
            if isinstance(max_radius, (int, float)):
                self.max_radius = float(max_radius)
            else:
                raise RuntimeError('Max radius must be an int or float')
        else:
            logger.warning('Running without a maximum radius! The proposal '
                           'process may get stuck if very large radii are '
                           'returned by the worst point.')
            self.max_radius = False

    def initialise(self):
        """
        Initialise the proposal class.

        This includes:
            * Setting up the rescaling
            * Verifying the rescaling is invertible
            * Intitialising the FlowModel
        """
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        self._x_dtype = False
        self._x_prime_dtype = False

        self.set_rescaling()
        self.verify_rescaling()
        if self.expansion_fraction and self.expansion_fraction is not None:
            logger.info('Overwritting fuzz factor with expansion fraction')
            self.fuzz = \
                (1 + self.expansion_fraction) ** (1 / self.rescaled_dims)
            logger.info(f'New fuzz factor: {self.fuzz}')
        self.flow_config['model_config']['n_inputs'] = self.rescaled_dims
        self.flow = FlowModel(config=self.flow_config, output=self.output)
        self.flow.initialise()
        self.populated = False
        self.initialised = True

    def update_poolsize_scale(self, acceptance):
        """
        Update poolsize given the current acceptance.

        Parameters
        ----------
        acceptance : float
            Current acceptance.
        """
        logger.debug(f'Updating poolsize with acceptance: {acceptance:.3f}')
        if not acceptance:
            logger.warning('Acceptance is zero, using maximum scale')
            self._poolsize_scale = self.max_poolize_scale
        else:
            self._poolsize_scale = 1.0 / acceptance
            if self._poolsize_scale > self.max_poolsize_scale:
                logger.warning(
                    'Poolsize scaling is greater than maximum value')
                self._poolsize_scale = self.max_poolsize_scale
            if self._poolsize_scale < 1.:
                self._poolsize_scale = 1.

    def set_boundary_inversion(self):
        """
        Setup boundary inversion
        """
        if self.boundary_inversion:
            if not self.rescale_parameters:
                raise RuntimeError('Boundary inversion requires rescaling')

            if (isinstance(self.boundary_inversion, list) and
                    not set(self.boundary_inversion).issubset(self.names)):
                raise RuntimeError(
                            'Boundaries are not in known parameters')
            elif isinstance(self.rescale_parameters, list):
                if (isinstance(self.boundary_inversion, list) and
                        not set(self.boundary_inversion).issubset(
                            self.rescale_parameters)):
                    raise RuntimeError(
                                'Boundaries are not in known parameters')

            if not isinstance(self.boundary_inversion, list):
                self.boundary_inversion = self.names.copy()

            logger.info('Appyling boundary inversion to: '
                        f'{self.boundary_inversion}')

            if self.inversion_type not in ('split', 'duplicate'):
                raise RuntimeError(
                        f'Unknown inversion type: {self.inversion_type}')

            self.rescale_bounds = [0, 1]
            self.update_bounds = True
            self._edges = {n: None for n in self.boundary_inversion}
            logger.info(f'Changing bounds to {self.rescale_bounds}')
        else:
            self.boundary_inversion = []

    def add_default_reparameterisations(self):
        """Add any reparmeterisations which are assumed by default"""
        logger.debug('No default reparameterisations')

    def get_reparameterisation(self, name):
        """Get the reparameterisation from the name"""
        return get_reparameterisation(name)

    def configure_reparameterisations(self, reparameterisations):
        _reparameterisations = copy.deepcopy(reparameterisations)
        logger.info(f'Adding reparameterisations from: {_reparameterisations}')
        self._reparameterisation = CombinedReparameterisation()

        if not isinstance(_reparameterisations, dict):
            raise TypeError('Reparameterisations must be a dictionary, '
                            f'receieved {type(_reparameterisations).__name__}')

        for k, config in _reparameterisations.items():
            if k in self.names:
                logger.debug(f'Found parameter {k} in model, '
                             'assuming it is a parameter')
                if isinstance(config, str) or config is None:
                    rc, default_config = self.get_reparameterisation(config)
                    default_config['parameters'] = k
                elif isinstance(config, dict):
                    if config.get('reparameterisation', None) is None:
                        raise RuntimeError(
                            f'No reparameterisation found for {k}. '
                            'Check inputs (and their spelling :)). '
                            f'Current keys: {list(config.keys())}')
                    rc, default_config = self.get_reparameterisation(
                        config['reparameterisation'])
                    config.pop('reparameterisation')

                    if config.get('parameters', False):
                        config['parameters'] += [k]
                    else:
                        default_config['parameters'] = k

                    default_config.update(config)
                else:
                    raise TypeError(
                        f'Unknown config type for: {k}. Expected str or dict, '
                        f'received instance of {type(config)}.')
            else:
                logger.debug(f'Assuming {k} is a reparameterisation')
                try:
                    rc, default_config = self.get_reparameterisation(k)
                    default_config.update(config)
                except ValueError:
                    raise RuntimeError(
                        f'{k} is not a parameter in the model or a known '
                        'reparameterisation')

            if not default_config.get('parameters', False):
                raise RuntimeError('No parameters key in the config! '
                                   'Check reparameterisations, setting logging'
                                   ' level to DEBUG can be helpful')

            if ('boundary_inversion' in default_config and
                    default_config['boundary_inversion']):
                self.boundary_inversion = True

            if isinstance(default_config['parameters'], list):
                prior_bounds = {p: self.model.bounds[p]
                                for p in default_config['parameters']}
            else:
                prior_bounds = \
                    {default_config['parameters']:
                     self.model.bounds[default_config['parameters']]}

            logger.debug(f'Adding {rc.__name__} with config {default_config}')
            r = rc(prior_bounds=prior_bounds, **default_config)
            self._reparameterisation.add_reparameterisations(r)

        self.add_default_reparameterisations()

        p = set(self.names) - set(self._reparameterisation.parameters)
        if p:
            logger.info(f'Assuming no rescaling for {p}')
            r = NullReparameterisation(parameters=list(p))
            self._reparameterisation.add_reparameterisations(r)

        if any(r._update_bounds for r in self._reparameterisation.values()):
            self.update_bounds = True
        else:
            self.update_bounds = False

        if self._reparameterisation.has_prime_prior:
            self.use_x_prime_prior = True
            self.x_prime_log_prior = self._reparameterisation.x_prime_log_prior
            logger.info('Using x prime prior')
        else:
            logger.info('Prime prior is disabled')
            if self._reparameterisation.requires_prime_prior:
                raise RuntimeError(
                    'One or more reparameterisations require use of the x '
                    'prime prior but it cannot be enabled with the current '
                    'settings.')

        self.rescale = self._rescale_w_reparameterisation
        self.inverse_rescale = \
            self._inverse_rescale_w_reparameterisation

        self.names = self._reparameterisation.parameters
        self.rescaled_names = self._reparameterisation.prime_parameters
        self.rescale_parameters = \
            list(set(self._reparameterisation.parameters)
                 - set(self._reparameterisation.prime_parameters))

    def set_rescaling(self):
        """
        Set function and parameter names for rescaling
        """
        self.names = self.model.names.copy()
        self.rescaled_names = self.names.copy()

        self.set_boundary_inversion()

        if self.model.reparameterisations is not None:
            self.configure_reparameterisations(self.model.reparameterisations)
            self.reparameterisations = self.model.reparameterisations
        elif self.reparameterisations is not None:
            self.configure_reparameterisations(self.reparameterisations)
        elif self.rescale_parameters:
            # if rescale is a list, there are the parameters to rescale
            # else all parameters are rescale
            if not isinstance(self.rescale_parameters, list):
                self.rescale_parameters = self.names.copy()

            for i, rn in enumerate(self.rescaled_names):
                if rn in self.rescale_parameters:
                    self.rescaled_names[i] += '_prime'

            self._min = {n: self.model.bounds[n][0] for n in self.model.names}
            self._max = {n: self.model.bounds[n][1] for n in self.model.names}
            self._rescale_factor = np.ptp(self.rescale_bounds)
            self._rescale_shift = self.rescale_bounds[0]

            self.rescale = self._rescale_to_bounds
            self.inverse_rescale = self._inverse_rescale_to_bounds
            logger.info(f'Set to rescale inputs to {self.rescale_bounds}')

            if self.update_bounds:
                logger.info(
                        'Rescaling will use min and max of current live points'
                        )
            else:
                logger.info('Rescaling will use model bounds')

        logger.info(f'x space parameters: {self.names}')
        logger.info(f'parameters to rescale {self.rescale_parameters}')
        logger.info(f'x prime space parameters: {self.rescaled_names}')

    def verify_rescaling(self):
        """
        Verify the rescaling functions are invertible
        """
        logger.info('Verifying rescaling functions')
        x = self.model.new_point(N=1000)
        for inversion in ['lower', 'upper', False, None]:
            self.check_state(x)
            logger.debug(f'Testing: {inversion}')
            x_prime, log_J = self.rescale(x, test=inversion)
            x_out, log_J_inv = self.inverse_rescale(x_prime)
            if x.size == x_out.size:
                for f in x.dtype.names:
                    if not np.allclose(x[f], x_out[f]):
                        raise RuntimeError(
                            f'Rescaling is not invertible for {f}')
                if not np.allclose(log_J, -log_J_inv):
                    raise RuntimeError('Rescaling Jacobian is not invertible')
            else:
                # ratio = x_out.size // x.size
                for f in x.dtype.names:
                    if not any([np.any(np.isclose(x[f], xo))
                                for xo in x_out[f]]):
                        raise RuntimeError(
                            'Duplicate samples must map to same input values. '
                            'Check the rescaling and inverse rescaling '
                            f'functions for {f}.')
                for f in x.dtype.names:
                    if not np.allclose(x[f], x_out[f][:x.size]):
                        raise RuntimeError(
                            f'Rescaling is not invertible for {f}')
                if not np.allclose(log_J, -log_J_inv):
                    raise RuntimeError('Rescaling Jacobian is not invertible')

        logger.info('Rescaling functions are invertible')

    def _rescale_w_reparameterisation(self, x, compute_radius=False, **kwargs):
        x_prime = np.zeros([x.size], dtype=self.x_prime_dtype)
        log_J = np.zeros(x_prime.size)

        if x.size == 1:
            x = np.array([x], dtype=x.dtype)

        x, x_prime, log_J = self._reparameterisation.reparameterise(
            x, x_prime, log_J, **kwargs)

        x_prime['logP'] = x['logP']
        x_prime['logL'] = x['logL']
        return x_prime, log_J

    def _inverse_rescale_w_reparameterisation(self, x_prime, **kwargs):
        x = np.zeros([x_prime.size], dtype=self.x_dtype)
        log_J = np.zeros(x.size)
        x, x_prime, log_J = self._reparameterisation.inverse_reparameterise(
            x, x_prime, log_J, **kwargs)

        x['logP'] = x_prime['logP']
        x['logL'] = x_prime['logL']
        return x, log_J

    def _rescale_to_bounds(self, x, compute_radius=False, test=None):
        """
        Rescale the inputs to specified bounds
        """
        x_prime = np.zeros([x.size], dtype=self.x_prime_dtype)
        log_J = np.zeros(x_prime.size)

        if x.size == 1:
            x = np.array([x], dtype=x.dtype)

        for n, rn in zip(self.names, self.rescaled_names):
            if n in self.rescale_parameters:
                x_prime[rn] = self._rescale_factor \
                             * ((x[n] - self._min[n])
                                / (self._max[n] - self._min[n])) \
                             + self._rescale_shift

                log_J += (-np.log(self._max[n] - self._min[n])
                          + np.log(self._rescale_factor))
                if n in self.boundary_inversion:
                    if self._edges[n] is None:
                        logger.debug('Determining edge')
                        self._edges[n] = detect_edge(
                            x_prime[rn],
                            test=test,
                            **self.detect_edges_kwargs
                            )
                    if self._edges[n]:
                        logger.debug(
                            f'Apply inversion for {n} to '
                            f'{self._edges[n]} bound'
                            )
                        if self._edges[n] == 'upper':
                            x_prime[rn] = 1 - x_prime[rn]
                        if (self.inversion_type == 'duplicate' or
                                compute_radius):
                            x_inv = x_prime.copy()
                            x_inv[rn] *= -1
                            x_prime = np.concatenate([x_prime, x_inv])
                            x = np.concatenate([x,  x])
                            log_J = np.concatenate([log_J, log_J])
                        else:
                            inv = np.random.choice(x_prime.size,
                                                   x_prime.size // 2,
                                                   replace=False)
                            x_prime[rn][inv] *= -1
                    else:
                        logger.debug(f'Not using inversion for {n}')
            else:
                x_prime[rn] = x[n]
        x_prime['logP'] = x['logP']
        x_prime['logL'] = x['logL']
        return x_prime, log_J

    def _inverse_rescale_to_bounds(self, x_prime):
        """
        Rescale the inputs from the prime space to the phyiscal space
        using the bounds specified
        """
        x = np.zeros([x_prime.size], dtype=self.x_dtype)
        log_J = np.zeros(x_prime.size)
        for n, rn in zip(self.names, self.rescaled_names):
            if n in self.rescale_parameters:
                if n in self.boundary_inversion:
                    inv = x_prime[rn] < 0.
                    x_prime[rn][~inv] = x_prime[rn][~inv]
                    x_prime[rn][inv] = -x_prime[rn][inv]

                    if self._edges[n] == 'upper':
                        x_prime[rn] = 1 - x_prime[rn]

                x[n] = (self._max[n] - self._min[n]) \
                    * (x_prime[rn] - self._rescale_shift) \
                    / self._rescale_factor + self._min[n]

                log_J += (np.log(self._max[n] - self._min[n])
                          - np.log(self._rescale_factor))
            else:
                x[n] = x_prime[rn]
        x['logP'] = x_prime['logP']
        x['logL'] = x_prime['logL']
        return x, log_J

    def rescale(self, x, compute_radius=False, **kwargs):
        """
        Rescale from the phyisical space to the primed physical space

        Parameters
        ----------
        x: array_like
            Array of live points to rescale
        compute_radius: bool (False)
            Used to indicate when rescaling is being used for computing the
            radius for population. This is important for rescaling that uses
            inversions.

        Returns
        -------
        array
            Array of rescaled values
        array
            Array of log det|J|
        """
        log_J = np.zeros(x.size)
        return x, log_J

    def inverse_rescale(self, x_prime, **kwargs):
        """
        Rescale from the primed physical space to the original physical
        space.

        Parameters
        ----------
        x_prime : array_like
            Array of live points to rescale.

        Returns
        -------
        array
            Array of rescaled values in the data space.
        array
            Array of log-Jacobian determinants.
        """
        log_J = np.zeros(x_prime.size)
        return x_prime, log_J

    def check_state(self, x):
        """
        Operations that need to checked before training. These include
        updating the bounds for rescaling and resetting the bounds for
        inversion.

        Parameters
        ----------
        x: array_like
            Array of training live points which can be used to set parameters
        """
        if self.boundary_inversion:
            logger.debug('Reseting inversion')
            if self._reparameterisation is not None:
                self._reparameterisation.reset_inversion()
            else:
                self._edges = {n: None for n in self.boundary_inversion}
        if self.update_bounds:
            logger.debug('Updating bounds')
            if self._reparameterisation is not None:
                self._reparameterisation.update_bounds(x)
            else:
                self._min = {n: np.min(x[n]) for n in self.model.names}
                self._max = {n: np.max(x[n]) for n in self.model.names}

    def train(self, x, plot=True):
        """
        Train the normalising flow given some of the live points.

        Parameters
        ----------
        x : structured_array
            Array of live points
        plot : {True, False, 'all'}
            Enable or disable plots for during training. By default the plots
            are only one-dimenensional histograms, `'all'` includes corner
            plots with samples, these are often a fwe MB in size so
            proceed with caution!
        """
        if (plot and self._plot_training) or self.save_training_data:
            block_output = \
                self.output + f'/training/block_{self.training_count}/'
        else:
            block_output = self.output

        if not os.path.exists(block_output):
            os.makedirs(block_output, exist_ok=True)

        if self.save_training_data:
            save_live_points(x, f'{block_output}/training_data.json')

        self.training_data = x.copy()
        self.check_state(self.training_data)

        if self._plot_training == 'all' and plot:
            plot_live_points(x, c='logL',
                             filename=block_output + 'x_samples.png')

        x_prime, log_J = self.rescale(x)

        self.training_data_prime = x_prime.copy()

        if self.rescale_parameters and self._plot_training == 'all' and plot:
            plot_live_points(x_prime, c='logL',
                             filename=block_output + 'x_prime_samples.png')
        # Convert to numpy array for training and remove likelihoods and priors
        # Since the names of parameters may have changes, pull names from flows
        x_prime_array = live_points_to_array(x_prime, self.rescaled_names)
        self.flow.train(x_prime_array,
                        output=block_output, plot=self._plot_training)

        if self._plot_training and plot:
            z_training_data, _ = self.forward_pass(self.training_data,
                                                   rescale=True)
            z_gen = np.random.randn(x.size, self.dims)

            fig = plt.figure()
            plt.hist(np.sqrt(np.sum(z_training_data ** 2, axis=1)), 'auto')
            plt.xlabel('Radius')
            fig.savefig(block_output + 'radial_dist.png')
            plt.close(fig)

            plot_1d_comparison(z_training_data, z_gen,
                               labels=['z_live_points', 'z_generated'],
                               convert_to_live_points=True,
                               filename=block_output + 'z_comparison.png')
            x_prime_gen, log_prob = self.backward_pass(z_gen, rescale=False)
            x_prime_gen['logL'] = log_prob
            x_gen, log_J = self.inverse_rescale(x_prime_gen)
            x_gen, log_J, x_prime_gen = \
                self.check_prior_bounds(x_gen, log_J, x_prime_gen)
            x_gen['logL'] += log_J
            if self._plot_training == 'all':
                plot_live_points(
                    x_prime_gen,
                    c='logL',
                    filename=block_output + 'x_prime_generated.png'
                    )
            if self.rescale_parameters:
                if self._plot_training == 'all':
                    plot_live_points(x_gen, c='logL',
                                     filename=block_output + 'x_generated.png')
                plot_1d_comparison(
                    x_prime, x_prime_gen, parameters=self.rescaled_names,
                    labels=['live points', 'generated'],
                    filename=block_output + 'x_prime_comparison.png')

            plot_1d_comparison(x, x_gen, parameters=self.model.names,
                               labels=['live points', 'generated'],
                               filename=block_output + 'x_comparison.png')

        self.populated = False
        self.training_count += 1

    def reset_model_weights(self, **kwargs):
        """
        Reset the flow weights.

        Parameters
        ----------
        kwargs :
            Keyword arguments passed to
            :meth:`nessai.flowmodel.FlowModel.reset_model`.
        """
        self.flow.reset_model(**kwargs)

    def check_prior_bounds(self, x, *args):
        """
        Return only values that are within the prior bounds

        Parameters
        ----------
        x: array_like
            Array of live points which will compared against prior bounds
        *args:
            Aditional arrays which correspond to the array of live points.
            Only those corresponding to points within the prior bounds
            are returned

        Returns
        -------
        out: tuple of arrays
            Array containing the subset of the orignal arrays which fall within
            the prior bounds
        """
        idx = np.array(list(((x[n] >= self.model.bounds[n][0])
                             & (x[n] <= self.model.bounds[n][1]))
                       for n in self.model.names)).T.all(1)
        out = (a[idx] for a in (x,) + args)
        return out

    def forward_pass(self, x, rescale=True, compute_radius=True):
        """
        Pass a vector of points through the model

        Parameters
        ----------
        x : array_like
            Live points to map to the latent space
        rescale : bool, optional (True)
            Apply rescaling function
        compute_radius : bool, optional (True)
            Flag parsed to rescaling for rescaling specific to radius
            computation

        Returns
        -------
        x : array_like
            Samples in the latent sapce
        log_prob : array_like
            Log probabilties corresponding to each sample (including the
            jacobian)
        """
        log_J = 0
        if rescale:
            x, log_J_rescale = self.rescale(x, compute_radius=compute_radius)
            log_J += log_J_rescale

        x = live_points_to_array(x, names=self.rescaled_names)

        if x.ndim == 1:
            x = x[np.newaxis, :]
        if x.shape[0] == 1:
            if self.clip:
                x = np.clip(x, *self.clip)
        z, log_prob = self.flow.forward_and_log_prob(x)

        return z, log_prob + log_J

    def backward_pass(self, z, rescale=True):
        """
        A backwards pass from the model (latent -> real)

        Parameters
        ----------
        z : array_like
            Structured array of points in the latent space
        rescale : bool, optional (True)
            Apply inverse rescaling function

        Returns
        -------
        x : array_like
            Samples in the latent sapce
        log_prob : array_like
            Log probabilties corresponding to each sample (including the
            Jacobian)
        """
        # Compute the log probability
        try:
            x, log_prob = self.flow.sample_and_log_prob(
                z=z, alt_dist=self.alt_dist)
        except AssertionError:
            return np.array([]), np.array([])

        valid = np.isfinite(log_prob)
        x, log_prob = x[valid], log_prob[valid]
        x = numpy_array_to_live_points(x.astype(DEFAULT_FLOAT_DTYPE),
                                       self.rescaled_names)
        # Apply rescaling in rescale=True
        if rescale:
            x, log_J = self.inverse_rescale(x)
            # Include Jacobian for the rescaling
            log_prob -= log_J
            x, log_prob = self.check_prior_bounds(x, log_prob)
        return x, log_prob

    def radius(self, z, log_q=None):
        """
        Calculate the radius of a latent point or set of latent points.
        If multiple points are parsed the maximum radius is returned.

        Parameters
        ----------
        z : :obj:`np.ndarray`
            Array of points in the latent space
        log_q : :obj:`np.ndarray`, optional (None)
            Array of correponding probabilities. If specified
            then probability of the maximum radius is also returned.

        Returns
        -------
        tuple of arrays
            Tuple of array with the maximum raidus and correspoding log_q
            if it was a specified input.
        """
        if log_q is not None:
            r = np.sqrt(np.sum(z ** 2., axis=-1))
            i = np.argmax(r)
            return r[i], log_q[i]
        else:
            return np.nanmax(np.sqrt(np.sum(z ** 2., axis=-1)))

    def log_prior(self, x):
        """
        Compute the prior probability using the user-defined model

        Parameters
        ----------
        x : structured_array
            Array of samples

        Returns
        -------
        array_like
            Array of log prior probabilities
        """
        if self._reparameterisation:
            return self.model.log_prior(x) \
                    + self._reparameterisation.log_prior(x)
        else:
            return self.model.log_prior(x)

    def x_prime_log_prior(self, x):
        """
        Compute the prior in the prime space

        Parameters
        ----------
        x : array
            Samples in the X-prime space.
        """
        raise RuntimeError('Prime prior is not implemented')

    def compute_weights(self, x, log_q):
        """
        Compute weights for the samples.

        Computes the log weights for rejection sampling sampling such that
        that the maximum log probability is zero.

        Also sets the fields `logP` and `logL`. Note `logL` is set as the
        proposal probability.

        Parameters
        ----------
        x :  structed_arrays
            Array of points
        log_q : array_like
            Array of log proposal probabilties.

        Returns
        -------
        array_like
            Log-weights for rejection sampling.
        """
        if self.use_x_prime_prior:
            x['logP'] = self.x_prime_log_prior(x)
        else:
            x['logP'] = self.log_prior(x)

        x['logL'] = log_q
        log_w = x['logP'] - log_q
        log_w -= np.max(log_w)
        return log_w

    def rejection_sampling(self, z, worst_q=None):
        """
        Perform rejection sampling.

        Coverts samples from the latent space and computes the corresponding
        weights. Then returns samples using standard rejection sampling.

        Parameters
        ----------
        z :  ndarray
            Samples from the latent space
        worst_q : float, optional
            Lower bound on the log-probability computed using the flow that
            is used to truncate new samples. Not recommended.

        Returns
        -------
        array_like
            Array of accepted latent samples.
        array_like
            Array of accepted samples in the X space.
        """
        x, log_q = self.backward_pass(z, rescale=not self.use_x_prime_prior)

        if not x.size:
            return x, log_q

        if self.truncate:
            cut = log_q >= worst_q
            x = x[cut]
            log_q = log_q[cut]

        # rescale given priors used initially, need for priors
        log_w = self.compute_weights(x, log_q)
        log_u = np.log(np.random.rand(x.shape[0]))
        indices = np.where(log_w >= log_u)[0]

        return z[indices], x[indices]

    def convert_to_samples(self, x, plot=True):
        """
        Convert the array to samples ready to be used.

        This removes are auxliary parameters, (e.g. auxiliary radial
        parameters) and ensures the prior is computed. These samples can
        be directly used in the nested sampler.

        Parameters
        ----------
        x : array_like
            Array of samples
        plot : bool, optional
            If true a 1d histogram for each parameter of the pool is plotted.
            This includes a comparison the live points used to train the
            current realisation of the flow.

        Returns
        -------
        array
            Structured array of samples.
        """
        if self.use_x_prime_prior:
            if self._plot_pool and plot:
                plot_1d_comparison(
                    self.training_data_prime, x,
                    labels=['live points', 'pool'],
                    filename=(f'{self.output}/pool_prime_'
                              + f'{self.populated_count}.png'))

            x, _ = self.inverse_rescale(x)
            x['logP'] = self.model.log_prior(x)
        return rfn.repack_fields(x[self.model.names + ['logP', 'logL']])

    def populate(self, worst_point, N=10000, plot=True, r=None):
        """
        Populate a pool of latent points given the current worst point.

        Parameters
        ----------
        worst_point : structured_array
            The current worst point used to compute the radius of the contour
            in the latent space.
        N : int, optional (10000)
            The total number of points to populate in the pool
        plot : {True, False, 'all'}
            Enable or disable plots for during training. By default the plots
            are only one-dimenensional histograms, `'all'` includes corner
            plots with samples, these are often a fwe MB in size so
            proceed with caution!
        """
        if r is not None:
            logger.info(f'Using user inputs for radius {r}')
            worst_q = None
        elif self.fixed_radius:
            r = self.fixed_radius
            worst_q = None
        else:
            logger.debug(f'Populating with worst point: {worst_point}')
            if self.compute_radius_with_all:
                logger.debug('Using previous live points to compute radius')
                worst_point = self.training_data
            worst_z, worst_q = self.forward_pass(worst_point,
                                                 rescale=True,
                                                 compute_radius=True)
            r, worst_q = self.radius(worst_z, worst_q)
            if self.max_radius and r > self.max_radius:
                r = self.max_radius
            if self.min_radius and r < self.min_radius:
                r = self.min_radius

        logger.info(f'Populating proposal with lantent radius: {r:.5}')
        self.r = r

        self.alt_dist = self.get_alt_distribution()

        if not self.keep_samples or not self.indices:
            self.x = np.empty(N,  dtype=self.population_dtype)
            self.x['logP'] = np.nan * np.ones(N)
            self.indices = []
            z_samples = np.empty([N, self.dims])

        proposed = 0
        accepted = 0
        percent = 0.1
        warn = True

        while accepted < N:
            z = self.draw_latent_prior(self.dims, r=self.r,
                                       N=self.drawsize, fuzz=self.fuzz,
                                       **self.draw_latent_kwargs)
            proposed += z.shape[0]

            z, x = self.rejection_sampling(z, worst_q)

            if not x.size:
                continue

            if warn:
                if x.size / self.drawsize < 0.01:
                    logger.warning(
                        'Rejection sampling accepted less than 1 percent of '
                        f'samples! ({x.size / self.drawsize})')
                    warn = False

            n = min(x.size, N - accepted)
            self.x[accepted:(accepted+n)] = x[:n]
            z_samples[accepted:(accepted+n), ...] = z[:n]
            accepted += n
            if accepted > percent * N:
                logger.info(f'Accepted {accepted} / {N} points, '
                            f'acceptance: {accepted/proposed:.4}')
                percent += 0.1

        self.samples = self.convert_to_samples(self.x, plot=plot)

        if self._plot_pool and plot:
            self.plot_pool(z_samples, self.samples)

        if self.check_acceptance:
            self.approx_acceptance.append(self.compute_acceptance(worst_q))
            logger.debug(
                f'Current approximate acceptance {self.approx_acceptance[-1]}')
            self.evaluate_likelihoods()
            self.acceptance.append(
                self.compute_acceptance(worst_point['logL']))
            logger.debug(f'Current acceptance {self.acceptance[-1]}')
        else:
            self.samples['logL'] = np.zeros(self.samples.size,
                                            dtype=self.samples['logL'].dtype)

        self.indices = np.random.permutation(self.samples.size).tolist()
        self.population_acceptance = self.x.size / proposed
        self.populated_count += 1
        self.populated = True
        self._checked_population = False
        logger.info(f'Proposal populated with {len(self.indices)} samples')
        logger.info(
            f'Overall proposal acceptance: {self.x.size / proposed:.4}')

    def get_alt_distribution(self):
        """
        Get a distribution for the latent prior used to draw samples.
        """
        if self.latent_prior in ['uniform_nsphere', 'uniform_nball']:
            return get_uniform_distribution(self.dims, self.r * self.fuzz,
                                            device=self.flow.device)
        elif self.latent_prior == 'truncated_gaussian':
            if 'var' in self.draw_latent_kwargs:
                return get_multivariate_normal(
                    self.dims, var=self.draw_latent_kwargs['var'],
                    device=self.flow.device)

    def compute_acceptance(self, logL):
        """
        Compute how many of the current pool have log-likelihoods greater
        than the specified log-likelihood using the current value in the
        `logL` field.

        Parameters
        ----------
        float : logL
            Log-likelihood to use as the lower bound

        Returns
        -------
        float :
            Acceptance defined on [0, 1]
        """
        return (self.samples['logL'] > logL).sum() / self.samples.size

    def draw(self, worst_point):
        """
        Draw a replacement point. The new point is independent of the worst
        point. The worst point is only used during population.

        Parameters
        ----------
        worst_point : structured_array
            The current worst point used to compute the radius of the contour
            in the latent space.

        Returns
        -------
        structured_array
            New live point
        """
        if not self.populated:
            self.populating = True
            if self.update_poolsize:
                self.update_poolsize_scale(self.ns_acceptance)
            st = datetime.datetime.now()
            while not self.populated:
                self.populate(worst_point, N=self.poolsize)
            self.population_time += (datetime.datetime.now() - st)
            self.populating = False
        # new sample is drawn randomly from proposed points
        # popping from right end is faster
        index = self.indices.pop()
        new_sample = self.samples[index]
        if not self.indices:
            self.populated = False
            logger.debug('Proposal pool is empty')
        # make live point and return
        return new_sample

    def plot_pool(self, z, x):
        """
        Plot the pool of points.

        Parameters
        ----------
        z : array_like
            Latent samples to plot
        x : array_like
            Corresponding samples to plot in the physical space.
        """
        if self._plot_pool == 'all':
            plot_live_points(
                x, c='logL',
                filename=f'{self.output}/pool_{self.populated_count}.png')
        else:
            plot_1d_comparison(
                self.training_data, x, labels=['live points', 'pool'],
                filename=f'{self.output}/pool_{self.populated_count}.png')

            z_tensor = torch.from_numpy(z).to(self.flow.device)
            with torch.no_grad():
                if self.alt_dist is not None:
                    log_p = self.alt_dist.log_prob(z_tensor).cpu().numpy()
                else:
                    log_p = self.flow.model.base_distribution_log_prob(
                        z_tensor).cpu().numpy()

            fig, axs = plt.subplots(3, 1, figsize=(3, 9))
            axs = axs.ravel()
            axs[0].hist(x['logL'], 20, histtype='step', label='log q')
            axs[1].hist(x['logL'] - log_p, 20, histtype='step',
                        label='log J')
            axs[2].hist(np.sqrt(np.sum(z ** 2, axis=1)), 20,
                        histtype='step', label='Latent radius')
            axs[0].set_xlabel('Log q')
            axs[1].set_xlabel('Log |J|')
            axs[2].set_xlabel('r')
            plt.tight_layout()
            fig.savefig(
                f'{self.output}/pool_{self.populated_count}_log_q.png')
            plt.close(fig)

    def resume(self, model, flow_config, weights_file=None):
        """
        Resume the proposal.

        The model and config are not stored so these must be provided.

        Parameters
        ----------
        model : :obj:`nessai.model.Model`
            User-defined model used.
        flow_config : dict
            Configuration dictionary for the flow.
        weights_files : str, optional
            Weights file to try and load. If not provided the proposal
            tries to load the last weighst file.
        """
        self.model = model
        self.flow_config = flow_config
        self._reparameterisation = None

        if self.mask is not None:
            if isinstance(self.mask, list):
                m = np.array(self.mask)
            self.flow_config['model_config']['kwargs']['mask'] = m

        self.initialise()

        if weights_file is None:
            weights_file = self.weights_file

        # Flow might have exited before any weights were saved.
        if weights_file is not None:
            if os.path.exists(weights_file):
                self.flow.reload_weights(weights_file)
        else:
            logger.warning('Could not reload weights for flow')

        if self.update_bounds:
            if self.training_data is not None:
                self.check_state(self.training_data)
            elif self.training_data is None and self.training_count:
                raise RuntimeError(
                    'Could not resume! Missing training data!')

    def test_draw(self):
        """
        Test the draw method to ensure it returns a sample in the correct
        format and the the log prior is computed.
        """
        logger.debug(f'Testing {self.__class__.__name__} draw method')

        test_point = self.model.new_point()
        self.populate(test_point, N=1, plot=False, r=1.0)
        new_point = self.draw(test_point)

        if new_point['logP'] != self.model.log_prior(new_point):
            raise RuntimeError('Log prior of new point is incorrect!')

        self.reset()

        logger.debug(f'{self.__class__.__name__} passed draw test')

    def reset(self):
        """Reset the proposal"""
        self.samples = None
        self.x = None
        self.populated = False
        self.populated_count = 0

    def __getstate__(self):
        state = self.__dict__.copy()
        state['initialised'] = False
        state['weights_file'] = state['flow'].weights_file

        # Mask may be generate via permutation, so must be saved
        if 'mask' in state['flow'].model_config['kwargs']:
            state['mask'] = state['flow'].model_config['kwargs']['mask']
        else:
            state['mask'] = None
        state['pool'] = None
        if state['populated'] and self.indices:
            state['resume_populated'] = True
        else:
            state['resumed_populated'] = False

        # user provides model and config for resume
        # flow can be reconstructed from resume
        del state['_reparameterisation']
        del state['model']
        del state['_flow_config']
        del state['flow']

        return state

    def __setstate__(self, state):
        self.__dict__ = state



# -*- coding: utf-8 -*-
"""
Test the AnglePair reparmeterisation
"""
import numpy as np
from numpy.testing import assert_equal
import pytest

from nessai.reparameterisations import AnglePair
from nessai.livepoint import get_dtype

angle_pairs = [(['ra', 'dec'], [[0, 2 * np.pi], [-np.pi / 2, np.pi / 2]]),
               (['az', 'zen'], [[0, 2 * np.pi], [0, np.pi]]),
               (['zen', 'az'], [[0, np.pi], [0, 2 * np.pi]])]


@pytest.fixture(params=angle_pairs, scope='function')
def angles(request):
    return request.param


@pytest.fixture(scope='function')
def assert_invertibility():
    def test_invertibility(reparam, angles, radial=None):

        n = list(angles.values())[0].size
        x = np.zeros([n], dtype=get_dtype(reparam.parameters))
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        for a in reparam.angles:
            x[a] = angles[a]
        if radial is not None:
            x[reparam.radial] = radial

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(
            x, x_prime, log_j)

        for a in reparam.angles:
            np.testing.assert_array_equal(x[a], x_re[a])
        if radial is not None:
            np.testing.assert_array_equal(x[reparam.radial],
                                          x_re[reparam.radial])

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = \
            reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

        for a in reparam.angles:
            np.testing.assert_array_almost_equal(x[a], x_inv[a])
        if radial is not None:
            np.testing.assert_array_almost_equal(x[reparam.radial],
                                                 x_inv[reparam.radial])

        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_almost_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


@pytest.mark.parametrize('parameters', ['x', ['w', 'x', 'y', 'z']])
def test_parameters_error(parameters):
    """
    Make sure reparameterisations fails with too many or too few parameters.
    """
    with pytest.raises(RuntimeError) as excinfo:
        AnglePair(parameters=parameters, prior_bounds=None)
    assert 'Must use a pair' in str(excinfo.value)


def test_two_angles(angles):
    """Test the reparmaterisation with just the angles"""
    parameters = angles[0]
    prior_bounds = {parameters[0]: angles[1][0],
                    parameters[1]: angles[1][1]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    if 'ra' in parameters:
        assert reparam.convention == 'ra-dec'

    if 'az' in parameters:
        assert reparam.convention == 'az-zen'

    # Make sure parameter[0] is always ra or azimuth
    assert_equal(reparam.prior_bounds[reparam.angles[0]],
                 np.array([0, 2 * np.pi]))

    assert reparam.chi is not False
    assert hasattr(reparam.chi, 'rvs')
    assert reparam.has_prime_prior is False

    m = '_'.join(parameters[:2])
    assert reparam.angles == parameters[:2]
    assert reparam.radial == (m + '_radial')
    assert reparam.x == (m + '_x')
    assert reparam.y == (m + '_y')
    assert reparam.z == (m + '_z')


def test_ra_dec(assert_invertibility):
    """Test the invertibility when using RA and Dec"""
    parameters = ['ra', 'dec']
    prior_bounds = {parameters[0]: [0, 2 * np.pi],
                    parameters[1]: [-np.pi / 2, np.pi / 2]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds,
                        convention='ra-dec', prior='isotropic')

    n = 100
    angles = {'ra': np.random.uniform(*prior_bounds['ra'], n),
              'dec': np.arcsin(np.random.uniform(-1, 1, n))}
    assert assert_invertibility(reparam, angles)


def test_azimuth_zenith(assert_invertibility):
    """Test the inverbility when using azimuth and zenith"""
    parameters = ['az', 'zen']
    prior_bounds = {parameters[0]: [0, 2 * np.pi],
                    parameters[1]: [0, np.pi]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds,
                        convention='az-zen', prior='isotropic')

    n = 100
    angles = {'az': np.random.uniform(*prior_bounds['az'], n),
              'zen': np.arccos(np.random.uniform(-1, 1, n))}
    assert reparam.parameters[:2] == list(angles.keys())
    assert assert_invertibility(reparam, angles)


def test_w_radial(assert_invertibility):
    """Test the reparameterisation with a radial parameter"""
    parameters = ['r', 'ra', 'dec']
    prior_bounds = {parameters[0]: [0, 5],
                    parameters[1]: [0, 2 * np.pi],
                    parameters[2]: [-np.pi / 2, np.pi / 2]}

    reparam = AnglePair(parameters=parameters, prior_bounds=prior_bounds)

    assert reparam.parameters == ['ra', 'dec', 'r']
    assert reparam.angles == ['ra', 'dec']
    assert reparam.chi is False

    n = 100
    angles = {'ra': np.random.uniform(*prior_bounds['ra'], n),
              'dec': np.arcsin(np.random.uniform(-1, 1, n))}

    radial = np.random.uniform(*prior_bounds['r'], n)

    assert assert_invertibility(reparam, angles, radial=radial)



# -*- coding: utf-8 -*-
"""
Utilities specific to the gw subpackage.
"""
import logging
import numpy as np
from scipy import stats, interpolate

from ..utils import (
    logit,
    sigmoid,
    rescale_zero_to_one,
    inverse_rescale_zero_to_one
)

logger = logging.getLogger(__name__)

try:
    import lalsimulation as lalsim
    from lal import MSUN_SI as m_sol
except ImportError:
    logger.debug(
        'Could not import LAL functions, running with reduced functionality')

try:
    from astropy import cosmology as cosmo
    import astropy.units as u
except ImportError:
    logger.debug(
        'Could not import astropy, running with reduced functionality')


def determine_rescaled_bounds(prior_min, prior_max, x_min, x_max, invert):
    """
    Determine the values of the prior min and max in the rescaled
    space.
    Parameters
    ----------
    prior_min : float
        Mininum of the prior
    prior_max : float
        Maximum of the prior
    x_min : float
        New minimum
    x_max : float
        New maximum
    invert : false or {'upper', 'lower', 'both'}
        Type of inversion
    """
    lower = (prior_min - x_min) / (x_max - x_min)
    upper = (prior_max - x_min) / (x_max - x_min)
    if not invert:
        return 2 * lower - 1, 2 * upper - 1
    elif invert == 'upper':
        return lower - 1, 1 - lower
    elif invert == 'lower':
        return -upper, upper
    elif invert == 'both':
        return -0.5, 1.5
    else:
        raise RuntimeError


def angle_to_cartesian(alpha, r=None, scale=1.0):
    """
    Decompose an angle into a real and imaginary part
    """
    alpha = np.asarray(alpha)
    rescaled_alpha = alpha * scale
    if r is None:
        r = stats.chi.rvs(2, size=alpha.size)
    elif any(r < 0):
        raise RuntimeError('Radius cannot be negative.')
    x = r * np.cos(rescaled_alpha)
    y = r * np.sin(rescaled_alpha)
    return x, y, np.log(r)


def cartesian_to_angle(x, y, scale=1.0, zero='centre'):
    """
    Reconstruct an angle given the real and imaginary part. Assume the angle
    is defined on [0, 2 pi] / scale.

    Parameters
    ----------
    x, y : array_like
        Cartesian coordinates
    scale : float, optional
        Rescaling factor used to rescale from [0, 2pi]
    zero : str, {centre, bound}
        Specifiy is zero should be the central value or lower bound
    """
    radius = np.sqrt(x ** 2. + y ** 2.)
    if zero == 'centre':
        angle = np.arctan2(y, x) / scale
    else:
        angle = (np.arctan2(y, x) + 2. * np.pi) % (2. * np.pi) / scale
    return angle, radius, -np.log(radius)


def ra_dec_to_cartesian(ra, dec, dL=None):
    """
    Convert right ascension, declination and (optional) luminosity distance
    (defined on [0, 1]) to Cartesian coordinates.

    Parameters
    ----------
    ra, dec: array_like
        Right ascension and declination
    dL: array_like, optional
        Corresponding luminosity distance defined on [0, 1]. If None (default)
        radial componment is drawn froma chi distribution with 3 degrees of
        freedom

    Returns
    -------
    x, y, z: array_like
        Cartesian coordinates
    log_J: array_like
        Determinant of the log-Jacobian
    """
    if dL is None:
        dL = stats.chi.rvs(3, size=ra.size)
    x = dL * np.cos(dec) * np.cos(ra)
    y = dL * np.cos(dec) * np.sin(ra)
    z = dL * np.sin(dec)
    return x, y, z, 2. * np.log(dL) + np.log(np.cos(dec))


def cartesian_to_ra_dec(x, y, z):
    """
    Reconstruct an angle given the real and imaginary part

    Parameters
    ----------
    x, y, z: array_like
        Three dimensional Cartesian coordinates

    Returns:
    ra, dec: array_like
        Right ascension and declination
    dl: array_like
        Luminosity distance
    log_J: array_like
        Determinant of the log-Jacobian
    """
    dL = np.sqrt(np.sum([x ** 2., y ** 2., z ** 2.], axis=0))
    dec = np.arctan2(z, np.sqrt(x ** 2. + y ** 2.0))
    ra = np.arctan2(y, x) % (2. * np.pi)
    return ra, dec, dL, - 2. * np.log(dL) - np.log(np.cos(dec))


def azimuth_zenith_to_cartesian(azimuth, zenith, dL=None):
    """
    Convert azimuth, zenith and (optional) luminosity distance
    (defined on [0, 1]) to Cartesian coordinates.

    Parameters
    ----------
    azimuth, zenith: array_like
        Azimuth and zenith
    dL: array_like, optional
        Corresponding luminosity distance defined on [0, 1]. If None (default)
        radial componment is drawn froma chi distribution with 3 degrees of
        freedom

    Returns
    -------
    x, y, z: array_like
        Cartesian coordinates
    log_J: array_like
        Determinant of the log-Jacobian
    """
    if dL is None:
        dL = stats.chi.rvs(3, size=azimuth.size)
    x = dL * np.sin(zenith) * np.cos(azimuth)
    y = dL * np.sin(zenith) * np.sin(azimuth)
    z = dL * np.cos(zenith)
    return x, y, z, 2. * np.log(dL) + np.log(np.sin(zenith))


def cartesian_to_azimuth_zenith(x, y, z):
    """
    Reconstruct an angle given the real and imaginary part

    Parameters
    ----------
    x, y, z: array_like
        Three dimensional Cartesian coordinates

    Returns:
    --------
    azimuth, zenith: array_like
        Azimuth and zenith
    dl: array_like
        Luminosity distance
    log_J: array_like
        Determinant of the log-Jacobian
    """
    dL = np.sqrt(np.sum([x ** 2., y ** 2., z ** 2.], axis=0))
    zenith = np.arctan2(np.sqrt(x ** 2. + y ** 2.0), z)
    azimuth = np.arctan2(y, x) % (2. * np.pi)
    return azimuth, zenith, dL, - 2. * np.log(dL) - np.log(np.sin(zenith))


def zero_one_to_cartesian(theta, mode='split'):
    """
    Convert a variable defined on [0,1] to an angle on [-pi, pi] and
    to Cartesian coordinates with a radius drawn from a chi distribution
    with two degrees of freedom. The lower bound is place at 0 and the upper
    bound at -pi/pi.

    Parameters
    ----------
    theta: array_like
        Array of values bound on [0, 1]

    Returns
    -------
    x, y: array_like
        Cartesian coordinates
    log_J: array_like
        Determinant of the log-Jacobian
    """
    theta = theta.copy()
    if mode == 'duplicate':
        theta = np.concatenate([theta, -theta])
    elif mode == 'split':
        neg = np.random.choice(theta.size, theta.size // 2, replace=False)
        theta[neg] *= -1
    elif mode == 'half':
        pass
    else:
        raise RuntimeError(f'Unknown mode: {mode}')
    return angle_to_cartesian(theta, scale=np.pi)


def cartesian_to_zero_one(x, y):
    """
    Convert Cartesian coordinates to a variable defined on [0, 1] and
    a corresponding radius.

    Parameters
    ----------
    x, y: array_like
        Cartesian coordinates

    Returns
    -------
    theta: array_like
        Variable defined on [0,1]
    radius: array_like
        Corresponding radius
    log_J: array_like
        Determinant of log-Jacobian
    """
    theta = np.abs(np.arctan2(y, x)) / np.pi
    radius = np.sqrt(x ** 2. + y ** 2.)
    return theta, radius, -np.log(radius)


def _spin_jacobian_to_precessing(sx, sy, sz):
    """Log Jacobian for transformation from spin vector to angles"""
    return - 0.5 * np.log(sx ** 2 + sy ** 2) \
           - 0.5 * np.log(sx ** 2 + sy ** 2 + sz ** 2)


def _spin_jacobian_from_precessing(a, theta):
    """Log jacobian for tranformation from precessing to spin vector"""
    return 2 * np.log(a) + np.log(np.sin(theta))


@np.vectorize
def transform_from_precessing_parameters(theta_jn, phi_jl, theta_1, theta_2,
                                         phi_12, a_1, a_2, m1, m2, f_ref,
                                         phase):
    f_ref = float(f_ref)
    m1 *= m_sol
    m2 *= m_sol
    iota, s1x, s1y, s1z, s2x, s2y, s2z = \
        lalsim.SimInspiralTransformPrecessingNewInitialConditions(
                theta_jn, phi_jl, theta_1, theta_2, phi_12, a_1, a_2, m1, m2,
                f_ref, phase)

    log_J = _spin_jacobian_from_precessing(a_1, theta_1)
    log_J += _spin_jacobian_from_precessing(a_2, theta_2)
    return iota, s1x, s1y, s1z, s2x, s2y, s2z, log_J


@np.vectorize
def transform_to_precessing_parameters(iota, s1x, s1y, s1z, s2x, s2y, s2z,
                                       m1, m2, f_ref, phase):
    f_ref = float(f_ref)
    theta_jn, phi_jl, theta_1, theta_2, phi_12, a_1, a_2 = \
        lalsim.SimInspiralTransformPrecessingWvf2PE(
            iota, s1x, s1y, s1z, s2x, s2y, s2z, m1, m2, f_ref, phase)

    log_J = _spin_jacobian_to_precessing(s1x, s1y, s1z)
    log_J += _spin_jacobian_to_precessing(s2x, s2y, s2z)
    return theta_jn, phi_jl, theta_1, theta_2, phi_12, a_1, a_2, log_J


def rescale_and_logit(x, xmin, xmax):
    x, log_J = rescale_zero_to_one(x, xmin, xmax)
    x, log_J_logit = logit(x)
    return x, log_J + log_J_logit


def rescale_and_sigmoid(x, xmin, xmax):
    x, log_J_sig = sigmoid(x)
    x, log_J = inverse_rescale_zero_to_one(x, xmin, xmax)
    return x, log_J + log_J_sig


class DistanceConverter:
    """Base object for converting from a distance parameter to a uniform
    parameter.
    """
    has_conversion = False
    has_Jacobian = False

    def to_uniform_parameter(self, d):
        raise NotImplementedError

    def from_uniform_parameter(self, d):
        raise NotImplementedError


class NullDistanceConverter(DistanceConverter):

    has_Jacobian = True

    def __init__(self, d_min=None, d_max=None, **kwargs):
        if kwargs:
            logger.warning(f'Kwargs {kwargs} will be ignored for distance')

    def to_uniform_parameter(self, d):
        return d, np.zeros_like(d)

    def from_uniform_parameter(self, d):
        return d. np.zeros_like(d)


class PowerLawConverter(DistanceConverter):
    """Convert from a distance parameter sample from a power law to a uniform
    parameter

    Assumes d proportional d^power.

    Parameters
    ----------
    power : float
        Power to use for the power-law.
    scale : float
        Factor used to rescale distance prior to convetering to the uniform
        parameter.
    """
    has_conversion = True
    has_jacobian = True

    def __init__(self, power=None, scale=1000.0, **kwargs):
        if power is None:
            raise RuntimeError(
                'Must specify the power to use in the power-law')
        self.power = power
        self.scale = scale
        self._power = self.power + 1

        if self._power == 3:
            self._f = np.cbrt
        elif self._power == 2:
            self._f = np.sqrt
        else:
            self._f = lambda x: x ** (1 / self._power)

    def _log_jacobian(self, d):
        return - self._power * np.log(self.scale) + np.log(self._power) + \
                (self._power - 1) * np.log(d)

    def _log_jacobian_inv(self, d):
        return np.log(self.scale) - np.log(self._power) + \
                (1 / self._power - 1) * np.log(d)

    def to_uniform_parameter(self, d):
        """Convert distance to a parameter with a uniform prior"""
        return (d / self.scale) ** (self._power), self._log_jacobian(d)

    def from_uniform_parameter(self, d):
        """Convert to distance from a parameter that has a uniform prior"""
        return self.scale * self._f(d), self._log_jacobian_inv(d)


class ComovingDistanceConverter(DistanceConverter):
    """
    Object to convert luminosity distance with a prior that is uniform
    in comoving volume to a parameter with a uniform prior.

    Parameters
    ----------
    d_min, d_max : float
        Minimum and maximum distances
    units : str, optional
        Units used for the distance, must be compatible with astropy units.
    cosomology : str, optional
        Cosmology used for conversion, must be compatible with astropy.
    scale : float, optional
        Rescaling applied to distance after converting to comoving distance.
    pad : float, optional
        Padding used for min and max of interpolation range:
        min = (1 - pad) * d_min and max = (1 + pad) * d_max
    n_interp : int, optional
        Length of vector used for generating the look up table.
    """
    has_conversion = True
    has_jacobian = False

    def __init__(self, d_min=None, d_max=None, units='Mpc',
                 cosmology='Planck15', scale=1000.0, pad=0.05, n_interp=500):
        self.units = u.Unit(units)
        # TODO: this needs to update with bilby
        self.cosmology = cosmo.Planck15
        self.scale = np.float64(scale)
        self.pad = pad
        self.n_interp = n_interp

        self.dl_min = (1 - self.pad) * d_min
        self.dl_max = (1 + self.pad) * d_max

        logger.debug(f'Min and max distances: [{self.dl_min}, {self.dl_max}]')

        self.dc_min = self.cosmology.comoving_distance(cosmo.z_at_value(
            self.cosmology.luminosity_distance, self.dl_min * self.units)
            ).value
        self.dc_max = self.cosmology.comoving_distance(cosmo.z_at_value(
            self.cosmology.luminosity_distance, self.dl_max * self.units)
            ).value

        logger.debug('Making distance look up table')

        dc_array = np.linspace(self.dc_min, self.dc_max, self.n_interp)
        dl_array = self.cosmology.luminosity_distance(
            [cosmo.z_at_value(self.cosmology.comoving_distance, d * self.units)
                for d in dc_array]).value

        self.interp_dc2dl = interpolate.splrep(dc_array, dl_array)
        self.interp_dl2dc = interpolate.splrep(dl_array, dc_array)

    def to_uniform_parameter(self, dl):
        """Convert luminosity distance to a parameter with a uniform prior"""
        return ((interpolate.splev(dl, self.interp_dl2dc, ext=3) /
                 self.scale) ** 3.,
                np.zeros_like(dl))

    def from_uniform_parameter(self, dc):
        """Convert comoving distance to a parameter with a uniform prior"""
        return (interpolate.splev(self.scale * np.cbrt(dc),
                                  self.interp_dc2dl, ext=3),
                np.zeros_like(dc))


def get_distance_converter(prior):
    """Get a distance converter from the type of prior.

    If the prior is unknown a null converter is returned that has the
    identity rescaling.
    """
    if prior == 'uniform-comoving-volume':
        return ComovingDistanceConverter
    if prior == 'power-law':
        return PowerLawConverter
    else:
        logger.info(f'Prior {prior} is not known for distance')
        return NullDistanceConverter



# -*- coding: utf-8 -*-
"""
Test the null reparameterisation
"""
import numpy as np
from numpy.testing import assert_array_equal

from nessai.reparameterisations import NullReparameterisation
from nessai.livepoint import numpy_array_to_live_points, get_dtype


def test_invertiblity():
    """Ensure the reparameterisation does not change the values"""
    reparam = NullReparameterisation(parameters=['x', 'y'])
    n = 100
    x = numpy_array_to_live_points(np.random.rand(n, 2), reparam.parameters)
    x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
    log_j = 0

    x_re, x_prime_re, log_j_re = reparam.reparameterise(x, x_prime, log_j)

    assert_array_equal(x, x_re)
    assert_array_equal(x, x_prime_re)

    x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))
    x_inv, x_prime_inv, log_j_inv = \
        reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

    assert_array_equal(x, x_inv)
    assert_array_equal(x, x_prime_inv)
    assert log_j == log_j_re == log_j_inv



# -*- coding: utf-8 -*-
"""
Test the Angle reparameterisation
"""
import numpy as np
import pytest

from nessai.reparameterisations import Angle
from nessai.livepoint import get_dtype

scales = [1.0, 2.0]


@pytest.fixture(params=scales, scope='function')
def scale(request):
    return request.param


@pytest.fixture(scope='function')
def assert_invertibility(model, n=100):
    def test_invertibility(reparam, angle, radial=None):

        x = np.zeros([n], dtype=get_dtype(reparam.parameters))
        x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
        log_j = 0

        x[reparam.angle] = angle
        if radial is not None:
            x[reparam.radial] = radial

        assert x.size == x_prime.size

        x_re, x_prime_re, log_j_re = reparam.reparameterise(
            x, x_prime, log_j)

        np.testing.assert_array_equal(x[reparam.angle], x_re[reparam.angle])
        if radial is not None:
            np.testing.assert_array_equal(x[reparam.radial],
                                          x_re[reparam.radial])

        x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

        x_inv, x_prime_inv, log_j_inv = \
            reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

        np.testing.assert_array_almost_equal(x[reparam.angle],
                                             x_inv[reparam.angle])
        if radial is not None:
            np.testing.assert_array_almost_equal(x[reparam.radial],
                                                 x_inv[reparam.radial])

        np.testing.assert_array_equal(x_prime_re, x_prime_inv)
        np.testing.assert_array_almost_equal(log_j_re, -log_j_inv)

        return True

    return test_invertibility


def test_angle_parameter():
    """Test init with just an angle parameter"""
    parameter = 'theta'
    prior_bounds = {parameter: np.array([0, 2 * np.pi])}
    reparam = Angle(parameters=parameter, prior_bounds=prior_bounds)

    assert reparam.chi is not False
    assert hasattr(reparam.chi, 'rvs')
    assert reparam._zero_bound is True
    assert reparam.has_prime_prior is False

    assert reparam.angle == parameter
    assert reparam.radial == (parameter + '_radial')
    assert reparam.radius == (parameter + '_radial')
    assert reparam.x == (parameter + '_x')
    assert reparam.y == (parameter + '_y')


def test_both_parameters():
    """Test init with just an angle and radial parameter"""
    parameters = ['theta', 'r']
    prior_bounds = {parameters[0]: np.array([0, 2 * np.pi]),
                    parameters[1]: np.array([0, 5])}

    reparam = Angle(parameters=parameters, prior_bounds=prior_bounds)

    assert reparam.chi is False
    assert reparam._zero_bound is True
    assert reparam.has_prime_prior is False

    assert reparam.angle == parameters[0]
    assert reparam.radial == parameters[1]


@pytest.mark.parametrize('angle_prior', [np.array([0, 2 * np.pi]),
                                         np.array([-np.pi, np.pi])])
def test_invertiblity_single_parameter(angle_prior, scale,
                                       assert_invertibility):
    """Test the inverbility when using just an angle"""
    n = 100
    parameter = 'theta'
    prior_bounds = {parameter: angle_prior / scale}
    reparam = Angle(parameters=parameter, prior_bounds=prior_bounds,
                    scale=scale)
    angle = np.random.uniform(*prior_bounds[parameter], n)
    assert assert_invertibility(reparam, angle, radial=None)


@pytest.mark.parametrize('angle_prior', [np.array([0, 2 * np.pi]),
                                         np.array([-np.pi, np.pi])])
def test_invertiblity_both_parameters(angle_prior,
                                      scale, assert_invertibility):
    """Test the inverbility when using just an angle"""
    n = 100
    parameters = ['theta', 'r']
    prior_bounds = {parameters[0]: angle_prior / scale,
                    parameters[1]: np.array([0, 5])}
    reparam = Angle(parameters=parameters, prior_bounds=prior_bounds,
                    scale=scale)
    angle = np.random.uniform(*prior_bounds[parameters[0]], n)
    radial = np.random.uniform(*prior_bounds[parameters[1]], n)
    assert assert_invertibility(reparam, angle, radial=radial)



# -*- coding: utf-8 -*-
"""
Functions and objects related to the main nested sampling algorithm.
"""
from collections import deque
import datetime
import logging
import os
import pickle

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from tqdm import tqdm

from .livepoint import get_dtype, DEFAULT_FLOAT_DTYPE
from .plot import plot_indices, plot_trace
from .posterior import log_integrate_log_trap
from .proposal import FlowProposal
from .utils import (
    safe_file_dump,
    compute_indices_ks_test,
    )

sns.set()
sns.set_style('ticks')

logger = logging.getLogger(__name__)


class _NSintegralState:
    """
    Stores the state of the nested sampling integrator
    """
    def __init__(self, nlive):
        self.nlive = nlive
        self.reset()

    def reset(self):
        """
        Reset the sampler to its initial state at logZ = -infinity
        """
        self.logZ = -np.inf
        self.oldZ = -np.inf
        self.logw = 0
        self.info = [0.]
        # Start with a dummy sample enclosing the whole prior
        self.logLs = [-np.inf]   # Likelihoods sampled
        self.log_vols = [0.0]    # Volumes enclosed by contours
        self.gradients = [0]

    def increment(self, logL, nlive=None):
        """
        Increment the state of the evidence integrator
        Simply uses rectangle rule for initial estimate
        """
        if(logL <= self.logLs[-1]):
            logger.warning('NS integrator received non-monotonic logL.'
                           f'{self.logLs[-1]:.5f} -> {logL:.5f}')
        if nlive is None:
            nlive = self.nlive
        oldZ = self.logZ
        logt = - 1.0 / nlive
        Wt = self.logw + logL + np.log1p(-np.exp(logt))
        self.logZ = np.logaddexp(self.logZ, Wt)
        # Update information estimate
        if np.isfinite(oldZ) and np.isfinite(self.logZ) and np.isfinite(logL):
            info = np.exp(Wt - self.logZ) * logL \
                  + np.exp(oldZ - self.logZ) \
                  * (self.info[-1] + oldZ) \
                  - self.logZ
            if np.isnan(info):
                info = 0
            self.info.append(info)

        # Update history
        self.logw += logt
        self.logLs.append(logL)
        self.log_vols.append(self.logw)
        self.gradients.append((self.logLs[-1] - self.logLs[-2])
                              / (self.log_vols[-1] - self.log_vols[-2]))

    def finalise(self):
        """
        Compute the final evidence with more accurate integrator
        Call at end of sampling run to refine estimate
        """
        # Trapezoidal rule
        self.logZ = log_integrate_log_trap(np.array(self.logLs),
                                           np.array(self.log_vols))
        return self.logZ

    def plot(self, filename):
        """
        Plot the logX vs logL
        """
        fig = plt.figure()
        plt.plot(self.log_vols, self.logLs)
        plt.title(f'logZ={self.logZ:.2f}'
                  f'H={self.info[-1] * np.log2(np.e):.2f} bits')
        plt.grid(which='both')
        plt.xlabel('log prior_volume')
        plt.ylabel('log likelihood')
        plt.xlim([self.log_vols[-1], self.log_vols[0]])
        plt.yscale('symlog')
        fig.savefig(filename)
        logger.info('Saved nested sampling plot as {0}'.format(filename))


class NestedSampler:
    """
    Nested Sampler class.
    Initialisation arguments:

    Parameters
    ----------
    model: :obj:`nessai.model.Model`
        User defined model
    nlive : int, optional
        Number of live points.
    output : str
        Path for output
    stopping : float, optional
        Stop when remaining samples wouldn't change logZ estimate by this much.
    max_iteration : int, optional
        Maximum number of iterations to run before force sampler to stop.
        If stopping criteria is met before max. is reached sampler will stop.
    checkpoint : bool, optional
        Boolean to toggle checkpointing, must be enable to resume sampler
    resume_file : str, optional
        If specified sampler will be resumed from this file. Still requieres
        correct model.
    seed : int, optional
        seed for the initialisation of the pseudorandom chain
    plot : bool (True)
        Boolean to toggle plotting
    proposal_plots : bool (True)
        Boolean to enable additional plots for the population stage of the
        sampler. Overwritten by plot.
    prior_sampling : bool (False)
        produce nlive samples from the prior.
    analytic_priors : bool (False)
        Boolean that indicates that the `new_point` method in the model
        draws directly from the priors meaning rejection sampling is not
        needed.
    maximum_uninformed : int (1000)
        Maximum number of iterations before forcing the sampler to switch to
        using the proposal method with the flow.
    uninformed_proposal : :obj:`nessai.proposal.Proposal`: (None)
        Class to use for inintial sampling before training the flow. If
        None RejectionProposal or AnalyticProposal are used depending if
        `analytic_priors` is False or True.
    uninformed_acceptance_threshold : float (None)
        Acceptance threshold for initialing sampling, if acceptance falls
        below this value sampler switches to flow-based proposal. If None
        then value is set to 10 times `acceptance_threshold`
    uninformed_proposal_kwargs : dict, ({})
        Dictionary of keyword argument to parase to the class use for
        the intial sampling when it is initialised.
    flow_class : :obj:`nessai.proposal.FlowProposal`
        Class to use for flow-based proposal method
    flow_config : dict ({})
        Dictionary used to configure instance of `nessai.flowmodel.FlowModel`,
        this includes configuring the normalising flow and the training.
    training_frequency : int (None)
        Number of iterations between re-training the flow. If None flow
        is only re-trained based on other criteria.
    train_on_empty : bool (True)
        If true the flow is retrained every time the proposal pool is
        empty. If false it is only training according to the other criteria.
    cooldown : int (100)
        Minimum number of iterations between training. Can be overridden if
        `train_on_empty=True` and the pool is empty.
    memory : int, False (False)
        Number of old live points to use in training. If False only the current
        live points are used.
    reset_weights : bool, int, (False)
        Boolean to toggle reseting the flow weights whenever re-training.
        If an interger is specified the flow is reset every nth time it is
        trained.
    reset_permuations: bool, int, (False)
        Boolean to toggle reseting the permuation layers in the flow whenever
        re-training. If an interger is specified the flow is reset every nth
        time it is trained.
    reset_acceptance : bool, (True)
        If true use mean acceptance of samples produced with current flow
        as a criteria for retraining
    retrain_acceptance : bool (False)
        Force the flow to be reset if the acceptance falls below the acceptance
        threshold. Requiers `reset_acceptance=True`
    acceptance_threshold : float (0.01)
        Threshold to determine if the flow should be retrained, will not
        retrain if cooldown is not satisfied.
    kwargs :
        Keyword arguments passed to the flow proposal class
    """

    def __init__(self, model, nlive=1000, output=None,
                 stopping=0.1,
                 max_iteration=None,
                 checkpointing=True,
                 checkpoint_on_training=False,
                 resume_file=None,
                 seed=None,
                 n_pool=None,
                 plot=True,
                 proposal_plots=True,
                 prior_sampling=False,
                 analytic_priors=False,
                 maximum_uninformed=1000,
                 uninformed_proposal=None,
                 uninformed_acceptance_threshold=None,
                 uninformed_proposal_kwargs=None,
                 flow_class=None,
                 flow_config=None,
                 training_frequency=None,
                 train_on_empty=True,
                 cooldown=100,
                 memory=False,
                 reset_weights=False,
                 reset_permutations=False,
                 retrain_acceptance=True,
                 reset_acceptance=False,
                 acceptance_threshold=0.01,
                 **kwargs):

        logger.info('Initialising nested sampler')

        self.info_enabled = logger.isEnabledFor(logging.INFO)

        model.verify_model()

        self.model = model
        self.nlive = nlive
        self.n_pool = n_pool
        self.live_points = None
        self.prior_sampling = prior_sampling
        self.setup_random_seed(seed)
        self.accepted = 0
        self.rejected = 1

        self.checkpointing = checkpointing
        self.checkpoint_on_training = checkpoint_on_training
        self.iteration = 0
        self.acceptance_history = deque(maxlen=(nlive // 10))
        self.mean_acceptance_history = []
        self.block_acceptance = 1.
        self.mean_block_acceptance = 1.
        self.block_iteration = 0
        self.retrain_acceptance = retrain_acceptance
        self.reset_acceptance = reset_acceptance

        self.insertion_indices = []
        self.rolling_p = []

        self.resumed = False
        self.tolerance = stopping
        self.condition = np.inf
        self.logLmin = -np.inf
        self.logLmax = -np.inf
        self.nested_samples = []
        self.logZ = None
        self.state = _NSintegralState(self.nlive)
        self.plot = plot
        self.resume_file = self.setup_output(output, resume_file)
        self.output = output

        # Timing
        self.training_time = datetime.timedelta()
        self.sampling_time = datetime.timedelta()
        self.sampling_start_time = datetime.datetime.now()

        # Resume flags
        self.completed_training = True
        self.finalised = False

        # History
        self.likelihood_evaluations = []
        self.training_iterations = []
        self.min_likelihood = []
        self.max_likelihood = []
        self.logZ_history = []
        self.dZ_history = []
        self.population_acceptance = []
        self.population_radii = []
        self.population_iterations = []
        self.checkpoint_iterations = []

        if max_iteration is None:
            self.max_iteration = np.inf
        else:
            self.max_iteration = max_iteration

        self.acceptance_threshold = acceptance_threshold

        self.train_on_empty = train_on_empty
        self.cooldown = cooldown
        self.memory = memory

        self.configure_flow_reset(reset_weights, reset_permutations)

        if training_frequency in [None, 'inf', 'None']:
            logger.warning('Proposal will only train when empty')
            self.training_frequency = np.inf
        else:
            self.training_frequency = training_frequency

        self.max_count = 0

        self.initialised = False

        if uninformed_proposal_kwargs is None:
            uninformed_proposal_kwargs = {}
        self.configure_uninformed_proposal(uninformed_proposal,
                                           analytic_priors,
                                           maximum_uninformed,
                                           uninformed_acceptance_threshold,
                                           **uninformed_proposal_kwargs)
        self.configure_flow_proposal(flow_class, flow_config, proposal_plots,
                                     **kwargs)

        # Uninformed proposal is used for prior sampling
        # If maximum uninformed is greater than 0, the it will be used for
        # another n interation or until it becomes inefficient

        self.store_live_points = False
        if self.store_live_points:
            self.live_points_dir = f'{self.output}/live_points/'
            os.makedirs(self.live_points_dir, exist_ok=True)
            self.replacement_points = []

    @property
    def log_evidence(self):
        return self.state.logZ

    @property
    def information(self):
        return self.state.info[-1]

    @property
    def likelihood_calls(self):
        return self.model.likelihood_evaluations

    @property
    def likelihood_evaluation_time(self):
        t = self._uninformed_proposal.logl_eval_time
        t += self._flow_proposal.logl_eval_time
        return t

    @property
    def proposal_population_time(self):
        t = self._uninformed_proposal.population_time
        t += self._flow_proposal.population_time
        return t

    @property
    def acceptance(self):
        return self.iteration / self.likelihood_calls

    @property
    def current_sampling_time(self):
        return self.sampling_time \
                + (datetime.datetime.now() - self.sampling_start_time)

    @property
    def last_updated(self):
        """Last time the normalising flow was retraining"""
        if self.training_iterations:
            return self.training_iterations[-1]
        else:
            return 0

    def configure_uninformed_proposal(self,
                                      uninformed_proposal,
                                      analytic_priors,
                                      maximum_uninformed,
                                      uninformed_acceptance_threshold,
                                      **kwargs):
        """
        Setup the uninformed proposal method (is NOT trained)

        Parameters
        ----------
        uninformed_proposal : None or obj
            Class to use for uninformed proposal
        analytic_priors : bool
            If True `AnalyticProposal` is used to directly sample from the
            priors rather than using rejection sampling.
        maximum_uninformed : {False, None, int, float}
            Maximum number of iterations before switching to FlowProposal.
            If None, no max is set. If False uninformed sampling is not used.
        uninformed_acceptance_threshold :  float or None:
            Threshold to use for uninformed proposal, once reached proposal
            method will switch. If None acceptance_threshold is used if
            greater than 0.1 else 10 x acceptance_threshold is used.
        kwargs
            Kwargs are passed to init method for uninformed proposal class
        """
        if maximum_uninformed is None:
            self.uninformed_sampling = True
            self.maximum_uninformed = np.inf
        elif not maximum_uninformed:
            self.uninformed_sampling = False
            self.maximum_uninformed = 0
        else:
            self.uninformed_sampling = True
            self.maximum_uninformed = maximum_uninformed

        if uninformed_acceptance_threshold is None:
            if self.acceptance_threshold < 0.1:
                self.uninformed_acceptance_threshold = \
                    10 * self.acceptance_threshold
            else:
                self.uninformed_acceptance_threshold = \
                    self.acceptance_threshold
        else:
            self.uninformed_acceptance_threshold = \
                uninformed_acceptance_threshold

        if uninformed_proposal is None:
            if analytic_priors:
                from .proposal import AnalyticProposal as uninformed_proposal
            else:
                from .proposal import RejectionProposal as uninformed_proposal
                kwargs['poolsize'] = self.nlive

        logger.debug(f'Using uninformed proposal: {uninformed_proposal}')
        logger.debug(f'Parsing kwargs to uninformed proposal: {kwargs}')
        self._uninformed_proposal = uninformed_proposal(
            self.model, n_pool=self.n_pool, **kwargs)

    def configure_flow_proposal(self, flow_class, flow_config, proposal_plots,
                                **kwargs):
        """
        Set up the flow-based proposal method

        Parameters
        ----------
        flow_class : None or obj
            Class to use for proposal. If None FlowProposal is used.
        flow_config : dict
            Configuration dictionary passed to the class.
        proposal_plots : bool or str
            Configuration of plottinmg in proposal class.
        **kwargs :
            Kwargs passed to init function.
        """
        proposal_output = self.output + '/proposal/'

        if not self.plot:
            proposal_plots = False

        if flow_class is not None:
            if isinstance(flow_class, str):
                if flow_class == 'GWFlowProposal':
                    from .gw.proposal import GWFlowProposal as flow_class
                elif flow_class == 'LegacyGWFlowProposal':
                    from .gw.proposal import LegacyGWFlowProposal as flow_class
                elif flow_class == 'FlowProposal':
                    flow_class = FlowProposal
                else:
                    raise RuntimeError(f'Unknown flow class: {flow_class}')
            elif not issubclass(flow_class, FlowProposal):
                raise RuntimeError('Flow class must be string or class that '
                                   'inherits from FlowProposal')
        else:
            flow_class = FlowProposal

        if kwargs.get('poolsize', None) is None:
            kwargs['poolsize'] = self.nlive

        logger.debug(f'Using flow class: {flow_class}')
        logger.info(f'Parsing kwargs to FlowProposal: {kwargs}')
        self._flow_proposal = flow_class(
            self.model, flow_config=flow_config, output=proposal_output,
            plot=proposal_plots, n_pool=self.n_pool, **kwargs)

    def setup_output(self, output, resume_file=None):
        """
        Set up the output folder

        Parameters
        ----------
        output : str
            Directory where the results will be stored
        resume_file : optional
            Specific file to use for checkpointing. If not specified the
            default is used (nested_sampler_resume.pkl)

        Returns
        -------
        resume_file : str
            File used for checkpointing
        """
        if not os.path.exists(output):
            os.makedirs(output, exist_ok=True)

        if resume_file is None:
            resume_file = os.path.join(output, "nested_sampler_resume.pkl")
        else:
            resume_file = os.path.join(output, resume_file)

        if self.plot:
            os.makedirs(output + '/diagnostics/', exist_ok=True)

        return resume_file

    def setup_random_seed(self, seed):
        """
        initialise the random seed
        """
        self.seed = seed
        if self.seed is not None:
            logger.debug(f'Setting random seed to {seed}')
            np.random.seed(seed=self.seed)
            torch.manual_seed(self.seed)

    def configure_flow_reset(self, reset_weights, reset_permutations):
        if isinstance(reset_weights, (int, float)):
            self.reset_weights = float(reset_weights)
        else:
            raise RuntimeError
        if isinstance(reset_permutations, (int, float)):
            self.reset_permutations = float(reset_permutations)
        else:
            raise RuntimeError

    def check_insertion_indices(self, rolling=True, filename=None):
        """
        Checking the distibution of the insertion indices either during
        the nested sampling run (rolling=True) or for the whole run
        (rolling=False).
        """
        if rolling:
            indices = self.insertion_indices[-self.nlive:]
        else:
            indices = self.insertion_indices

        D, p = compute_indices_ks_test(indices, self.nlive)

        if p is not None:
            if rolling:
                logger.warning(f'Rolling KS test: D={D:.4}, p-value={p:.4}')
                self.rolling_p.append(p)
            else:
                logger.warning(f'Final KS test: D={D:.4}, p-value={p:.4}')

        if filename is not None:
            np.savetxt(os.path.join(self.output, filename),
                       self.insertion_indices, newline='\n', delimiter=' ')

    def log_likelihood(self, x):
        """
        Wrapper for the model likelihood so evaluations are counted
        """
        return self.model.log_likelihood(x)

    def yield_sample(self, oldparam):
        """
        Draw points and applying rejection sampling
        """
        while True:
            counter = 0
            while True:
                counter += 1
                newparam = self.proposal.draw(oldparam.copy())

                # Prior is computed in the proposal
                if newparam['logP'] != -np.inf:
                    if not newparam['logL']:
                        newparam['logL'] = \
                            self.model.evaluate_log_likelihood(newparam)
                    if newparam['logL'] > self.logLmin:
                        self.logLmax = max(self.logLmax, newparam['logL'])
                        oldparam = newparam.copy()
                        break
                # Only here if proposed and then empty
                # This returns the old point and allows for a training check
                if not self.proposal.populated:
                    break
            yield counter, oldparam

    def insert_live_point(self, live_point):
        """
        Insert a live point
        """
        # This is the index including the current worst point, so final index
        # is one less, otherwise index=0 would never be possible
        index = np.searchsorted(self.live_points['logL'], live_point['logL'])
        self.live_points[:index - 1] = self.live_points[1:index]
        self.live_points[index - 1] = live_point
        return index - 1

    def consume_sample(self):
        """
        Replace a sample for single thread
        """
        worst = self.live_points[0].copy()
        self.logLmin = worst['logL']
        self.state.increment(worst['logL'])
        self.nested_samples.append(worst)

        self.condition = np.logaddexp(self.state.logZ,
                                      self.logLmax
                                      - self.iteration / float(self.nlive)) \
            - self.state.logZ

        # Replace the points we just consumed with the next acceptable ones
        # Make sure we are mixing the chains
        self.iteration += 1
        self.block_iteration += 1
        count = 0

        while(True):
            c, proposed = next(self.yield_sample(worst))
            count += c

            if proposed['logL'] > self.logLmin:
                # Assuming point was proposed
                # replace worst point with new one
                index = self.insert_live_point(proposed)
                self.insertion_indices.append(index)
                self.accepted += 1
                self.block_acceptance += 1 / count
                self.acceptance_history.append(1 / count)
                break
            else:
                # Only get here if the yield sample returns worse point
                # which can only happen if the pool is empty
                self.rejected += 1
                self.check_state()
                # if retrained whilst proposing a sample then update the
                # iteration count since will be zero otherwise
                if not self.block_iteration:
                    self.block_iteration += 1

        self.mean_block_acceptance = self.block_acceptance \
            / self.block_iteration

        if self.info_enabled:
            logger.info(f"{self.iteration:5d}: n: {count:3d} "
                        f"b_acc: {self.mean_block_acceptance:.3f} "
                        f"H: {self.state.info[-1]:.2f} "
                        f"logL: {self.logLmin:.5f} --> {proposed['logL']:.5f} "
                        f"dZ: {self.condition:.3f} "
                        f"logZ: {self.state.logZ:.3f} "
                        f"+/- {np.sqrt(self.state.info[-1] / self.nlive):.3f} "
                        f"logLmax: {self.logLmax:.2f}")

    def populate_live_points(self):
        """
        Initialise the pool of live points.
        """
        i = 0
        live_points = np.empty(self.nlive,
                               dtype=get_dtype(self.model.names,
                                               DEFAULT_FLOAT_DTYPE))

        with tqdm(total=self.nlive, desc='Drawing live points') as pbar:
            while i < self.nlive:
                while i < self.nlive:
                    count, live_point = next(
                            self.yield_sample(self.model.new_point()))
                    if np.isnan(live_point['logL']):
                        logger.warning('Likelihood function returned NaN for '
                                       'live_points ' + str(live_points[i]))
                        logger.warning(
                            'You may want to check your likelihood function')
                    if (live_point['logP'] != -np.inf and
                            live_point['logL'] != -np.inf):
                        live_points[i] = live_point
                        i += 1
                        pbar.update()
                        break

        self.live_points = np.sort(live_points, order='logL')
        if self.store_live_points:
            np.savetxt(self.live_points_dir + '/intial_live_points.dat',
                       self.live_points,
                       header='\t'.join(self.live_points.dtype.names))

    def initialise(self, live_points=True):
        """
        Initialise the nested sampler

        Parameters
        ----------
        live_points : bool, optional (True)
            If true and there are no live points, new live points are
            drawn using `populate_live_points` else all other initialisation
            steps are complete but live points remain empty.
        """
        flags = [False] * 3
        if not self._flow_proposal.initialised:
            self._flow_proposal.initialise()
            flags[0] = True

        if not self._uninformed_proposal.initialised:
            self._uninformed_proposal.initialise()
            flags[1] = True

        if self.iteration < self.maximum_uninformed:
            self.proposal = self._uninformed_proposal
        else:
            self.proposal = self._flow_proposal

        self.proposal.configure_pool()

        if live_points and self.live_points is None:
            self.populate_live_points()
            flags[2] = True

        if self.condition > self.tolerance:
            self.finalised = False

        if all(flags):
            self.initialised = True

    @property
    def mean_acceptance(self):
        """
        Mean acceptance of the last nlive // 10 points
        """
        return np.mean(self.acceptance_history)

    def check_proposal_switch(self):
        """
        Check if the proposal should be switch from uninformed to
        flowproposal given the current state.

        Returns
        -------
        bool
            Flag to indicated if proposal was switched
        """
        if ((self.mean_acceptance < self.uninformed_acceptance_threshold)
                or (self.iteration >= self.maximum_uninformed)):
            logger.warning('Switching to FlowProposal')
            # Make sure the pool is closed
            if self.proposal.pool is not None:
                self.proposal.close_pool()
            self.proposal = self._flow_proposal
            if self.proposal.n_pool is not None:
                self.proposal.configure_pool()
            self.proposal.ns_acceptance = self.mean_block_acceptance
            self.uninformed_sampling = False
            return True
        # If using uninformed sampling, don't check training
        else:
            return False

    def check_training(self):
        """
        Check if the normalising flow should be trained

        Checks that can force training:
            - Training was previously stopped before completion
            - The pool is empty and the proposal was not in the process
              of populating when stopped.
        Checks that cannot force training is still on cooldown:
            - Acceptance falls below threshold and `retrain_acceptance` is
              true
            - The number of iterations since last training is equal to the
              training frequency

        Returns
        -------
        train : bool
            Try to train if true
        force : bool
            Force the training irrespective of cooldown
        """
        if not self.completed_training:
            logger.debug('Training flow (resume)')
            return True, True
        elif (not self.proposal.populated and
                self.train_on_empty and
                not self.proposal.populating):
            logger.debug('Training flow (proposal empty)')
            return True, True
        elif (self.mean_block_acceptance < self.acceptance_threshold and
                self.retrain_acceptance):
            logger.debug('Training flow (acceptance)')
            return True, False
        elif (self.iteration - self.last_updated) == self.training_frequency:
            logger.debug('Training flow (iteration)')
            return True, False
        else:
            return False, False

    def check_flow_model_reset(self):
        """
        Check if the normalising flow model should be reset
        """
        if (self.reset_acceptance
                and self.mean_block_acceptance < self.acceptance_threshold):
            self.proposal.reset_model_weights(weights=True, permutations=True)
            return

        if (self.reset_weights and
                not (self.proposal.training_count % self.reset_weights)):
            self.proposal.reset_model_weights(weights=True)

        if (self.reset_permutations and
                not (self.proposal.training_count % self.reset_permutations)):
            self.proposal.reset_model_weights(weights=False, permutations=True)

    def train_proposal(self, force=False):
        """
        Try to trin the proposal. Proposal will not train if cooldown is not
        exceeded unless force is True.

        Parameters
        ----------
        force : bool
            Override training checks
        """
        if (self.iteration - self.last_updated < self.cooldown and not force):
            logger.debug('Not training, still cooling down!')
        else:
            self.completed_training = False
            self.check_flow_model_reset()

            training_data = self.live_points.copy()
            if self.memory and (len(self.nested_samples) >= self.memory):
                training_data = np.concatenate([
                    training_data, self.nested_samples[-self.memory:].copy()])

            st = datetime.datetime.now()
            self.proposal.train(training_data)
            self.training_time += (datetime.datetime.now() - st)
            self.training_iterations.append(self.iteration)

            self.block_iteration = 0
            self.block_acceptance = 0.
            self.completed_training = True
            if self.checkpoint_on_training:
                self.checkpoint(periodic=True)

    def check_state(self, force=False):
        """
        Check if state should be updated prior to drawing a new sample

        Force will overide the cooldown mechanism.
        """

        if self.uninformed_sampling:
            if self.check_proposal_switch():
                force = True
            else:
                return
        # General overide
        train = False
        if force:
            train = True
            logger.debug('Training flow (force)')
        elif not train:
            train, force = self.check_training()

        if train or force:
            self.train_proposal(force=force)

    def plot_state(self):
        """
        Produce plots with the current state of the nested sampling run.
        Plots are saved to the output directory specifed at initialisation.
        """

        fig, ax = plt.subplots(6, 1, sharex=True, figsize=(12, 12))
        ax = ax.ravel()
        it = (np.arange(len(self.min_likelihood))) * (self.nlive // 10)
        it[-1] = self.iteration

        colours = ['#4575b4', '#d73027', '#fad117']

        ls = ['-', '--', ':']

        for t in self.training_iterations:
            for a in ax:
                a.axvline(t, ls='-', color='lightgrey')

        if not self.train_on_empty:
            for p in self.population_iterations:
                for a in ax:
                    a.axvline(p, ls='-', color='tab:orange')

        for i in self.checkpoint_iterations:
            for a in ax:
                a.axvline(i, ls='-', color='tab:pink')

        ax[0].plot(it, self.min_likelihood, label='Min logL',
                   c=colours[0], ls=ls[0])
        ax[0].plot(it, self.max_likelihood, label='Max logL',
                   c=colours[1], ls=ls[1])
        ax[0].set_ylabel('logL')
        ax[0].legend(frameon=False)

        g = np.min([len(self.state.gradients), self.iteration])
        ax[1].plot(np.arange(g), np.abs(self.state.gradients[:g]),
                   c=colours[0], label='Gradient')
        ax[1].set_ylabel(r'$|d\log L/d \log X|$')
        ax[1].set_yscale('log')

        ax[2].plot(it, self.likelihood_evaluations, c=colours[0], ls=ls[0],
                   label='Evalutions')
        ax[2].set_ylabel('logL evaluations')

        ax[3].plot(it, self.logZ_history, label='logZ', c=colours[0], ls=ls[0])
        ax[3].set_ylabel('logZ')
        ax[3].legend(frameon=False)

        ax_dz = plt.twinx(ax[3])
        ax_dz.plot(it, self.dZ_history, label='dZ', c=colours[1], ls=ls[1])
        ax_dz.set_ylabel('dZ')
        handles, labels = ax[3].get_legend_handles_labels()
        handles_dz, labels_dz = ax_dz.get_legend_handles_labels()
        ax[3].legend(handles + handles_dz, labels + labels_dz, frameon=False)

        ax[4].plot(it, self.mean_acceptance_history, c=colours[0],
                   label='Proposal')
        ax[4].plot(self.population_iterations, self.population_acceptance,
                   c=colours[1], ls=ls[1], label='Population')
        ax[4].set_ylabel('Acceptance')
        ax[4].set_ylim((-0.1, 1.1))
        handles, labels = ax[4].get_legend_handles_labels()

        ax_r = plt.twinx(ax[4])
        ax_r.plot(self.population_iterations, self.population_radii,
                  label='Radius', color=colours[2], ls=ls[2])
        ax_r.set_ylabel('Population radius')
        handles_r, labels_r = ax_r.get_legend_handles_labels()
        ax[4].legend(handles + handles_r, labels + labels_r, frameon=False)

        if len(self.rolling_p):
            it = (np.arange(len(self.rolling_p)) + 1) * self.nlive
            ax[5].plot(it, self.rolling_p, 'o', c=colours[0], label='p-value')
        ax[5].set_ylabel('p-value')
        ax[5].set_ylim([-0.1, 1.1])

        ax[-1].set_xlabel('Iteration')

        fig.suptitle(f'Sampling time: {self.current_sampling_time}',
                     fontsize=16)

        fig.tight_layout()
        fig.subplots_adjust(top=0.95)
        fig.savefig(f'{self.output}/state.png')
        plt.close(fig)

    def plot_trace(self):
        """
        Make trace plots for the nested samples
        """
        if self.nested_samples:
            plot_trace(self.state.log_vols[1:], self.nested_samples,
                       filename=f'{self.output}/trace.png')
        else:
            logger.warning('Could not produce trace plot. No nested samples!')

    def update_state(self, force=False):
        """
        Update state after replacing a live point
        """
        # Check if acceptance is not None, this indicates the proposal
        # was populated
        if not self.proposal._checked_population:
            self.population_acceptance.append(
                self.proposal.population_acceptance)
            self.population_radii.append(self.proposal.r)
            self.population_iterations.append(self.iteration)
            self.proposal._checked_population = True

        if not (self.iteration % (self.nlive // 10)) or force:
            self.likelihood_evaluations.append(
                    self.model.likelihood_evaluations)
            self.min_likelihood.append(self.logLmin)
            self.max_likelihood.append(self.logLmax)
            self.logZ_history.append(self.state.logZ)
            self.dZ_history.append(self.condition)
            self.mean_acceptance_history.append(self.mean_acceptance)

        if not (self.iteration % self.nlive) or force:
            logger.warning(
                f"it: {self.iteration:5d}: "
                f"n eval: {self.likelihood_calls} "
                f"H: {self.state.info[-1]:.2f} "
                f"dZ: {self.condition:.3f} logZ: {self.state.logZ:.3f} "
                f"+/- {np.sqrt(self.state.info[-1] / self.nlive):.3f} "
                f"logLmax: {self.logLmax:.2f}")
            self.checkpoint(periodic=True)
            if not force:
                self.check_insertion_indices()
                if self.plot:
                    plot_indices(self.insertion_indices[-self.nlive:],
                                 self.nlive,
                                 plot_breakdown=False,
                                 filename=(f'{self.output}/diagnostics/'
                                           'insertion_indices_'
                                           f'{self.iteration}.png'))

            if self.plot:
                self.plot_state()
                self.plot_trace()

            if self.uninformed_sampling:
                self.block_acceptance = 0.
                self.block_iteration = 0

        self.proposal.ns_acceptance = self.mean_block_acceptance

    def checkpoint(self, periodic=False):
        """
        Checkpoint the classes internal state

        Parameters
        ----------
        periodic : bool
            Indicates if the checkpoint is regular periodic checkpointing
            or forced by a signal
        """
        if not periodic:
            self.checkpoint_iterations += [self.iteration]
        self.sampling_time += \
            (datetime.datetime.now() - self.sampling_start_time)
        logger.critical('Checkpointing nested sampling')
        safe_file_dump(self, self.resume_file, pickle, save_existing=True)
        self.sampling_start_time = datetime.datetime.now()

    def check_resume(self):
        """
        Check the normalising flow is correctly configured is the sampler
        was resumed.
        """
        if self.resumed:
            # If pool is populated reset the flag since it is set to
            # false during initialisation
            if hasattr(self._flow_proposal, 'resume_populated'):
                if (self._flow_proposal.resume_populated and
                        self._flow_proposal.indices):
                    self._flow_proposal.populated = True
                    logger.info('Resumed with populated pool')

            self.resumed = False

    def finalise(self):
        """
        Finalise things after sampling
        """
        logger.info('Finalising')
        for i, p in enumerate(self.live_points):
            self.state.increment(p['logL'], nlive=self.nlive-i)
            self.nested_samples.append(p)

        # Refine evidence estimate
        self.update_state(force=True)
        self.state.finalise()
        # output the chain and evidence
        self.finalised = True

    def nested_sampling_loop(self):
        """
        Main nested sampling loop
        """
        self.sampling_start_time = datetime.datetime.now()
        if not self.initialised:
            self.initialise(live_points=True)

        if self.prior_sampling:
            for i in range(self.nlive):
                self.nested_samples = self.params.copy()
            return

        self.check_resume()

        if self.iteration:
            self.update_state()

        logger.critical('Starting nested sampling loop')

        while self.condition > self.tolerance:

            self.check_state()

            self.consume_sample()

            self.update_state()

            if self.iteration >= self.max_iteration:
                break

        if self.proposal.pool is not None:
            self.proposal.close_pool()

        # final adjustments
        # avoid repeating final adjustments if resuming a completed run.
        if not self.finalised and (self.condition <= self.tolerance):
            self.finalise()

        logger.critical(f'Final evidence: {self.state.logZ:.3f} +/- '
                        f'{np.sqrt(self.state.info[-1] / self.nlive):.3f}')
        logger.critical('Information: {0:.2f}'.format(self.state.info[-1]))

        self.check_insertion_indices(rolling=False)

        # This includes updating the total sampling time
        self.checkpoint(periodic=True)

        logger.info(f'Total sampling time: {self.sampling_time}')
        logger.info(f'Total training time: {self.training_time}')
        logger.info(f'Total population time: {self.proposal_population_time}')
        logger.info(
            f'Total likelihood evaluations: {self.likelihood_calls:3d}')
        if self.proposal.logl_eval_time.total_seconds():
            logger.info(
                'Time spent evaluating likelihood: '
                f'{self.likelihood_evaluation_time}')

        return self.state.logZ, np.array(self.nested_samples)

    @classmethod
    def resume(cls, filename, model, flow_config={}, weights_file=None):
        """
        Resumes the interrupted state from a checkpoint pickle file.

        Parameters
        ----------
        filename : str
            Pickle pickle to resume from
        model : :obj:`nessai.model.Model`
            User-defined model
        flow_config : dict, optional
            Dictionary for configuring the flow
        weights_file : str, optional
            Weights files to use in place of the weights file stored in the
            pickle file.

        Returns
        -------
        obj
            Instance of NestedSampler
        """
        logger.critical('Resuming NestedSampler from ' + filename)
        with open(filename, 'rb') as f:
            obj = pickle.load(f)
        model.likelihood_evaluations += obj.likelihood_evaluations[-1]
        obj.model = model
        obj._uninformed_proposal.resume(model)
        obj._flow_proposal.resume(model, flow_config, weights_file)

        obj.resumed = True
        return obj

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['model']
        return state

    def __setstate__(self, state):
        self.__dict__ = state



import pytest

from nessai.posterior import (
        logsubexp
        )


def test_logsubexp_negative():
    """
    Test behaviour of logsubexp for x < y
    """
    with pytest.raises(Exception):
        logsubexp(1, 2)



# -*- coding: utf-8 -*-
"""
Test the combined reparameterisation class.
"""
import numpy as np

from nessai.reparameterisations import (
    CombinedReparameterisation,
    RescaleToBounds
    )
from nessai.livepoint import get_dtype


def test_init_none():
    c = CombinedReparameterisation()
    assert c.parameters == []


def test_init_w_reparam():
    c = CombinedReparameterisation(RescaleToBounds('x', [0, 1]))
    assert c.parameters == ['x']
    assert c.prime_parameters == ['x_prime']


def test_add_single_reparameterisations():
    """Test the core functionailty of adding reparameterisations"""
    r = RescaleToBounds(parameters='x', prior_bounds=[0, 1])
    c = CombinedReparameterisation()
    c.add_reparameterisations(r)
    assert c.parameters == ['x']
    assert c.has_prime_prior is False


def test_add_multiple_reparameterisations(model):
    """
    Test adding multiple reparameterisations and using the reparameterisation.
    """
    r = [RescaleToBounds(parameters='x', prior_bounds=model.bounds['x'],
                         prior='uniform'),
         RescaleToBounds(parameters='y', prior_bounds=model.bounds['y'],
                         prior='uniform')]
    reparam = CombinedReparameterisation()
    reparam.add_reparameterisations(r)

    assert reparam.parameters == ['x', 'y']
    assert reparam.has_prime_prior is True

    n = 100
    x = model.new_point(N=n)
    x_prime = np.zeros([n], dtype=get_dtype(reparam.prime_parameters))
    log_j = 0

    x_re, x_prime_re, log_j_re = reparam.reparameterise(
        x, x_prime, log_j)

    assert reparam.x_prime_log_prior(x_prime_re) is not None

    np.testing.assert_array_equal(x, x_re)

    x_in = np.zeros([n], dtype=get_dtype(reparam.parameters))

    x_inv, x_prime_inv, log_j_inv = \
        reparam.inverse_reparameterise(x_in, x_prime_re, log_j)

    np.testing.assert_array_equal(x, x_inv)
    np.testing.assert_array_equal(x_prime_re, x_prime_inv)
    np.testing.assert_array_equal(log_j_re, -log_j_inv)



import pytest
import torch
import numpy as np

from nessai.flows import (
    FlexibleRealNVP,
    MaskedAutoregressiveFlow,
    NeuralSplineFlow
    )

flows = [FlexibleRealNVP, MaskedAutoregressiveFlow, NeuralSplineFlow]


@pytest.fixture()
def data_dim():
    return 2


@pytest.fixture()
def n():
    return 1000


@pytest.fixture()
def x(n, data_dim):
    return torch.randn(n, data_dim).float()


@pytest.fixture()
def z(n, data_dim):
    return torch.randn(n, data_dim).float()


@pytest.fixture(params=flows)
def flow(request, data_dim):
    return request.param(data_dim, 8, 2, 2).eval()


def test_forward(flow, x, n, data_dim):
    """Test the foward pass of the flow"""
    with torch.no_grad():
        z, _ = flow.forward(x)
    assert z.shape == (n, data_dim)


def test_inverse(flow, z, n, data_dim):
    """Test the inverse method"""
    with torch.no_grad():
        x, _ = flow.inverse(z)
    assert x.shape == (n, data_dim)


def test_sample(flow, n, data_dim):
    """Test the sample method"""
    x = flow.sample(n)
    assert x.shape == (n, data_dim)


def test_log_prob(flow, x, n):
    """Test the log prob method"""
    with torch.no_grad():
        log_prob = flow.log_prob(x)
    assert log_prob.shape == (n,)


def test_base_distribution_log_prob(flow, z, n):
    """Test the bast distribution"""
    with torch.no_grad():
        log_prob = flow.base_distribution_log_prob(z)
    assert log_prob.shape == (n,)


def test_forward_and_log_prob(flow, x, n, data_dim):
    """
    Test the forward and log prob method.

    Tests to ensure method runs and that it agrees with using forward
    and log_prob seperately
    """
    with torch.no_grad():
        z, log_prob = flow.forward_and_log_prob(x)
        z_target, _ = flow.forward(x)
        log_prob_target = flow.log_prob(x)
    np.testing.assert_array_equal(z.numpy(), z_target.numpy())
    np.testing.assert_array_equal(log_prob.numpy(), log_prob_target.numpy())


@pytest.mark.flaky(run=10)
def test_sample_and_log_prob(flow, n, data_dim):
    """
    Assert that samples are drawn with correct shape and that the log
    prob is correct.
    """
    with torch.no_grad():
        x, log_prob = flow.sample_and_log_prob(n)
        log_prob_target = flow.log_prob(x)
    assert x.shape == (n, data_dim)
    np.testing.assert_array_almost_equal(
        log_prob.numpy(), log_prob_target.numpy(), decimal=5)


def test_invertibility(flow, x):
    """Test to ensure flows are invertible"""
    with torch.no_grad():
        z, log_J = flow.forward(x)
        x_out, log_J_out = flow.inverse(z)

    np.testing.assert_array_almost_equal(x.numpy(), x_out.numpy())
    np.testing.assert_array_almost_equal(
        log_J.numpy(), -log_J_out.numpy(), decimal=5)



# -*- coding: utf-8 -*-
"""
Various utilies for implementeding normalising flows.
"""
import logging
import numpy as np
import torch
import torch.nn.functional as F

from nflows.transforms.normalization import BatchNorm
from nflows.transforms.lu import LULinear
from nflows.transforms.permutations import RandomPermutation
from nflows.nn.nets import MLP

from .realnvp import FlexibleRealNVP
from .maf import MaskedAutoregressiveFlow
from .nsf import NeuralSplineFlow
from .distributions import MultivariateNormal

logger = logging.getLogger(__name__)


def silu(x):
    """
    SiLU (Sigmoid-weighted Linear Unit) activation function.

    Also known as swish.

    Elfwing et al 2017: https://arxiv.org/abs/1702.03118v3
    """
    return torch.mul(x, torch.sigmoid(x))


def setup_model(config):
    """
    Setup the flow form a configuration dictionary.
    """
    kwargs = {}
    flows = {'realnvp': FlexibleRealNVP, 'maf': MaskedAutoregressiveFlow,
             'frealnvp': FlexibleRealNVP, 'spline': NeuralSplineFlow}
    activations = {'relu': F.relu, 'tanh': F.tanh, 'swish': silu, 'silu': silu}

    k = config.get('kwargs', None)
    if k is not None:
        if 'activation' in k and isinstance(k['activation'], str):
            try:
                k['activation'] = activations[k['activation']]
            except KeyError as e:
                raise RuntimeError(f'Unknown activation function {e}')

        if 'var' in k and 'distribution' not in k:
            k['distribution'] = MultivariateNormal([config['n_inputs']],
                                                   var=k['var'])
            k.pop('var')

        kwargs.update(k)

    if not isinstance(config['n_inputs'], int):
        raise TypeError('Number of inputs (n_inputs) must be an int')

    fc = config.get('flow', None)
    ftype = config.get('ftype', None)
    if fc is not None:
        model = fc(config['n_inputs'], config['n_neurons'], config['n_blocks'],
                   config['n_layers'], **kwargs)
    elif ftype is not None:
        if ftype.lower() not in flows:
            raise RuntimeError(f'Unknown flow type: {ftype}. Choose from:'
                               f'{flows.keys()}')
        if (('mask' in kwargs and kwargs['mask'] is not None) or
                ('net' in kwargs and kwargs['net'] is not None) and
                ftype not in ['realnvp', 'frealnvp']):
            raise RuntimeError('Custom masks and networks are only '
                               'supported for RealNVP')

        model = flows[ftype.lower()](config['n_inputs'], config['n_neurons'],
                                     config['n_blocks'], config['n_layers'],
                                     **kwargs)
    else:
        raise RuntimeError("Must specify either 'flow' or 'ftype'.")

    if 'device_tag' in config:
        if isinstance(config['device_tag'], str):
            device = torch.device(config['device_tag'])

        try:
            model.to(device)
        except RuntimeError as e:
            device = torch.device('cpu')
            logger.warning("Could not send the normailising flow to the "
                           f"specified device {config['device']} send to CPU "
                           f"instead. Error raised: {e}")
    logger.debug('Flow model:')
    logger.debug(model)

    model.device = device

    return model, device


def reset_weights(module):
    """Reset parameters of a given module in place.

    Uses the ``reset_parameters`` method from ``torch.nn.Module``

    Also checks the following modules from nflows

    - nflows.transforms.normalization.BatchNorm

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    if hasattr(module, 'reset_parameters'):
        module.reset_parameters()
    elif isinstance(module, BatchNorm):
        # nflows BatchNorm does not have a weight reset, so must
        # be done manually
        constant = np.log(np.exp(1 - module.eps) - 1)
        module.unconstrained_weight.data.fill_(constant)
        module.bias.data.zero_()
        module.running_mean.zero_()
        module.running_var.fill_(1)


def reset_permutations(module):
    """Resets permutations and linear transforms for a given module in place.

    Resets using theor original initialisation method. This needed since they
    do not have a ``reset_parameters`` method.

    Parameters
    ----------
    module : :obj:`torch.nn.Module`
        Module to reset
    """
    if isinstance(module, LULinear):
        module.cache.invalidate()
        module._initialize(identity_init=True)
    elif isinstance(module, RandomPermutation):
        module._permutation = torch.randperm(len(module._permutation))


class CustomMLP(MLP):
    """
    MLP which handles additional kwargs that are supplied by some
    flow models
    """
    def __init__(self, *args, **kwargs):
        super(CustomMLP, self).__init__(*args, **kwargs)

    def forward(self, inputs, *args, **kwargs):
        """Forward method that allows for kwargs such as context"""
        if inputs.shape[1:] != self._in_shape:
            raise ValueError(
                "Expected inputs of shape {}, got {}.".format(
                    self._in_shape, inputs.shape[1:]
                )
            )

        inputs = inputs.reshape(-1, np.prod(self._in_shape))
        outputs = self._input_layer(inputs)
        outputs = self._activation(outputs)

        for hidden_layer in self._hidden_layers:
            outputs = hidden_layer(outputs)
            outputs = self._activation(outputs)

        outputs = self._output_layer(outputs)
        if self._activate_output:
            outputs = self._activation(outputs)
        outputs = outputs.reshape(-1, *self._out_shape)

        return outputs



import matplotlib.pyplot as plt
import numpy as np
import pytest

from nessai import plot


@pytest.fixture()
def live_points(model):
    """Set of live points"""
    return model.new_point(N=100)


@pytest.fixture()
def nested_samples(live_points):
    return np.sort(live_points, order='logL')


@pytest.mark.parametrize('bounds', [None, True])
def test_plot_live_points_bounds(live_points, bounds, model):
    """Test generating a plot for a set of live points."""
    if bounds:
        bounds = model.bounds
    plot.plot_live_points(live_points, bounds=bounds)


@pytest.mark.parametrize('c', [None, 'x'])
def test_plot_live_points_hue(live_points, c, model):
    """Test generating a plot for a set of live points with a hue."""
    plot.plot_live_points(live_points, c=c)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_live_points_save(live_points, save, model, tmpdir):
    """Test generating a plot for a set of live points and saving it."""
    if save:
        filename = tmpdir + 'lp.png'
    else:
        filename = None
    plot.plot_live_points(live_points, filename=filename)
    plt.close()


@pytest.mark.parametrize('parameters', [None, ['x', 'y']])
def test_plot_1d_comparison_unstructured(parameters):
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(100, 2)
    plot.plot_1d_comparison(live_points, convert_to_live_points=True,
                            parameters=parameters)
    plt.close()


def test_plot_1d_comparison_unstructured_missing_flag():
    """Test plotting live points in arrays are not structured."""
    live_points = np.random.randn(100, 2)
    with pytest.raises(RuntimeError) as excinfo:
        plot.plot_1d_comparison(live_points, convert_to_live_points=False)

    assert 'not structured array' in str(excinfo.value)


@pytest.mark.parametrize('parameters', [None, ['x', 'y']])
def test_plot_1d_comparison_parameters(parameters, model):
    """Test generating a 1d comparison plot"""
    l1 = model.new_point(N=100)
    l2 = model.new_point(N=100)
    plot.plot_1d_comparison(l1, l2, parameters=parameters)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_1d_comparison_save(save, model, tmpdir):
    """Test generating a 1d comparison plot"""
    l1 = model.new_point(N=100)
    l2 = model.new_point(N=100)
    if save:
        filename = tmpdir + 'comp.png'
    else:
        filename = None
    plot.plot_1d_comparison(l1, l2, filename=filename)
    plt.close()


@pytest.mark.parametrize('labels', [None, True])
def test_plot_1d_comparison_labels(labels, model):
    """Test generating a 1d comparison plot"""
    l1 = model.new_point(N=100)
    l2 = model.new_point(N=100)
    if labels:
        labels = model.names
    plot.plot_1d_comparison(l1, l2, labels=labels)
    plt.close()


@pytest.mark.parametrize('bounds', [None, True])
def test_plot_1d_comparison_bounds(bounds, model):
    """Test generating a 1d comparison plot"""
    l1 = model.new_point(N=100)
    l2 = model.new_point(N=100)
    if bounds:
        bounds = model.bounds
    plot.plot_1d_comparison(l1, l2, bounds=bounds)
    plt.close()


@pytest.mark.parametrize('plot_breakdown', [False, True])
def test_plot_indices_breakdown(plot_breakdown):
    """Test plotting insertion indices with and without breakdown"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    plot.plot_indices(indices, nlive=nlive, plot_breakdown=plot_breakdown)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_indices_save(save, tmpdir):
    """Test plotting insertion indices with and without saving"""
    nlive = 100
    indices = np.random.randint(0, nlive, 1000)
    if save:
        filename = tmpdir + 'indices.png'
    else:
        filename = None
    plot.plot_indices(indices, nlive=nlive, filename=filename)
    plt.close()


def test_plot_indices_no_indices():
    """Test to ensure plotting does not fail if indices are empty"""
    plot.plot_indices([], nlive=100)
    plt.close()


@pytest.mark.parametrize('sym', [False, True])
def test_plot_loss(sym):
    """Test function for plotting the loss"""
    epoch = 10
    if sym:
        history = dict(loss=-np.arange(epoch),
                       val_loss=-np.arange(epoch))
    else:
        history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    plot.plot_loss(epoch, history)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_plot_loss_save(save, tmpdir):
    """Test function for plotting the loss"""
    epoch = 10
    history = dict(loss=np.arange(epoch), val_loss=np.arange(epoch))
    if save:
        filename = tmpdir + 'loss.png'
    else:
        filename = None
    plot.plot_loss(epoch, history, filename=filename)
    plt.close()


@pytest.mark.parametrize('save', [False, True])
def test_trace_plot_save(nested_samples, save, tmpdir):
    """Test trace plot generation saving."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    if save:
        filename = tmpdir + 'trace.png'
    else:
        filename = None
    plot.plot_trace(log_x, nested_samples, filename=filename)
    plt.close()


@pytest.mark.parametrize('labels', [None, ['x', 'y']])
def test_trace_plot_labels(nested_samples, labels, tmpdir):
    """Test trace plot generation with labels."""
    log_x = np.linspace(-10, 0, nested_samples.size)
    plot.plot_trace(log_x, nested_samples, labels=labels)
    plt.close()


def test_trace_plot_labels_error(nested_samples):
    """Test to ensure error is raised if labels are incompatible"""
    log_x = np.linspace(-10, 0, nested_samples.size)
    with pytest.raises(RuntimeError) as excinfo:
        plot.plot_trace(log_x, nested_samples, labels=['1', '2', '3'])

    assert 'Missing labels' in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test the base reparameterisation.
"""
import numpy as np
from numpy.testing import assert_equal
import pytest

from nessai.reparameterisations import Reparameterisation


@pytest.mark.parametrize('name', ['x1', ['x1']])
@pytest.mark.parametrize('prior_bounds', [[0, 1], (0, 1), {'x1': [0, 1]}])
def test_init(name, prior_bounds):
    """Test the init method with the allowed types of inputs"""
    reparam = Reparameterisation(parameters=name, prior_bounds=prior_bounds)
    assert reparam.parameters == ['x1']
    assert reparam.prime_parameters == ['x1_prime']
    assert_equal(reparam.prior_bounds, {'x1': np.array([0, 1])})


def test_parameters_error():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters={'x': [0, 1]})
    assert 'Parameters must be a str or list' in str(excinfo.value)


def test_missing_bounds():
    with pytest.raises(RuntimeError) as excinfo:
        Reparameterisation(parameters=['x', 'y'], prior_bounds={'x': [0, 1]})
    assert 'Mismatch' in str(excinfo.value)


def test_incorrect_bounds_type():
    with pytest.raises(TypeError) as excinfo:
        Reparameterisation(parameters=['x', 'y'], prior_bounds=1)
    assert 'Prior bounds must be' in str(excinfo.value)


def test_methods_not_implemented():
    """Test to ensure class fails if user does not define the methods"""
    reparam = Reparameterisation(parameters='x', prior_bounds=[0, 1])

    with pytest.raises(NotImplementedError):
        reparam.reparameterise(None, None, None)

    with pytest.raises(NotImplementedError):
        reparam.inverse_reparameterise(None, None, None)



# -*- coding: utf-8 -*-
"""
Base objects for implementing normalising flows.
"""
from torch.nn import Module


class BaseFlow(Module):
    """
    Base class for all normalising flows.

    If implementing flows using distributions and transforms see NFlow.
    """
    def forward(self, x, context=None):
        """
        Apply the forward transformation and return samples in the
        latent space and the log-Jacobian determinant.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log determinants of the Jacobian of the forward
            transformation
        """
        raise NotImplementedError()

    def inverse(self, z, context=None):
        """
        Apply the inverse transformation and return samples in the
        data space and the log-Jacobian determinant.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        :obj:`torch.Tensor`
            Tensor of log determinants of the Jacobian of the forward
            transformation
        """
        raise NotImplementedError()

    def sample(self, n, context=None):
        """
        Generate n samples in the data space

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        """
        raise NotImplementedError()

    def log_prob(self, x, context=None):
        """
        Compute the log probability for a set of samples in the data space

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()

    def base_distribution_log_prob(self, z, context=None):
        """
        Computes the log probability of samples in the latent for
        the base distribution in the flow.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of log probabilities of the latent samples
        """
        raise NotImplementedError()

    def forward_and_log_prob(self, x, context=None):
        """
        Apply the forward transformation and compute the log probability
        of each sample

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()

    def sample_and_log_prob(self, n, context=None):
        """
        Generates samples from the flow, together with their log probabilities
        in the data space ``log p(x) = log p(z) + log|J|``.

        For flows, this is more efficient that calling ``sample`` and
        ``log_prob`` separately.

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the data space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        raise NotImplementedError()


class NFlow(BaseFlow):
    """
    Base class for flow objects implemented according to outline in nflows.
    These take an instance of a transform and a distribution from nflows.

    This replaces `Flow` from nflows. It removes the context and includes
    additional methods which are called in FlowModel.

    Parameters
    ----------
    transform : :obj: `nflows.transforms.Transform`
        Object that applys the transformation, must have`forward` and
        `inverse` methods. See nflows for more details.
    distribution : :obj: `nflows.distributions.Distribution`
        Object the serves as the base distribution used when sampling
        and computing the log probrability. Must have `log_prob` and
        `sample` methods. See nflows for details
    """
    def __init__(self, transform, distribution):
        super().__init__()

        for method in ['forward', 'inverse']:
            if not hasattr(transform, method):
                raise RuntimeError(
                    f'Transform does not have `{method}` method')

        for method in ['log_prob', 'sample']:
            if not hasattr(distribution, method):
                raise RuntimeError(
                    f'Distribution does not have `{method}` method')

        self._transform = transform
        self._distribution = distribution

    def forward(self, x, context=None):
        """
        Apply the forward transformation and return samples in the latent
        space and log-Jacobian determinant.
        """
        return self._transform.forward(x)

    def inverse(self, z, context=None):
        """
        Apply the inverse transformation and return samples in the
        data space and log-Jacobian determinant (not log probability).
        """
        return self._transform.inverse(z)

    def sample(self, num_samples, context=None):
        """
        Produces N samples in the data space by drawing from the base
        distribution and the applying the inverse transform.

        Does NOT need to be specified by the user
        """
        noise = self._distribution.sample(num_samples)

        samples, _ = self._transform.inverse(noise)

        return samples

    def log_prob(self, inputs, context=None):
        """
        Computes the log probability of the inputs samples by apply the
        transform.

        Does NOT need to specified by the user
        """
        noise, logabsdet = self._transform(inputs)
        log_prob = self._distribution.log_prob(noise)
        return log_prob + logabsdet

    def base_distribution_log_prob(self, z, context=None):
        """
        Computes the log probability of samples in the latent for
        the base distribution in the flow.
        """
        return self._distribution.log_prob(z)

    def forward_and_log_prob(self, x, context=None):
        """
        Apply the forward transformation and compute the log probability
        of each sample

        Returns
        -------
        :obj:`torch.Tensor`
            Tensor of samples in the latent space
        :obj:`torch.Tensor`
            Tensor of log probabilities of the samples
        """
        z, log_J = self.forward(x)
        log_prob = self.base_distribution_log_prob(z)
        return z, log_prob + log_J

    def sample_and_log_prob(self, N, context=None):
        """
        Generates samples from the flow, together with their log probabilities
        in the data space ``log p(x) = log p(z) + log|J|``.

        For flows, this is more efficient that calling ``sample`` and
        ``log_prob`` separately.
        """
        z, log_prob = self._distribution.sample_and_log_prob(N)

        samples, logabsdet = self._transform.inverse(z)

        return samples, log_prob - logabsdet



# -*- coding: utf-8 -*-
"""
Proposal method for initial sampling when priors are not analytical.
"""
import datetime

import numpy as np

from .base import Proposal


class RejectionProposal(Proposal):
    """
    Object for rejection sampling from the priors.

    Relies on :meth:`nessai.model.Model.new_point`.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model
    poolsize : int, optional
        Number of new samples to store in the pool.
    """
    def __init__(self, model, poolsize=1000, **kwargs):
        super(RejectionProposal, self).__init__(model, **kwargs)
        self._poolsize = poolsize
        self.populated = False
        self._checked_population = True
        self.population_acceptance = None

    @property
    def poolsize(self):
        """Poolsize used for drawing new samples in batches."""
        return self._poolsize

    def draw_proposal(self):
        """Draw a signal new point"""
        return self.model.new_point(N=self.poolsize)

    def log_proposal(self, x):
        """
        Log proposal probability. Calls \
                :meth:`nessai.model.Model.new_point_log_prob`

        Parameters
        ----------
        x : structured_array
            Array of new points
        """
        return self.model.new_point_log_prob(x)

    def compute_weights(self, x):
        """
        Get weights for the samples.

        Computes the log weights for rejection sampling sampling such that
        that the maximum log probability is zero.

        Parameters
        ----------
        x :  structed_arrays
            Array of points
        """
        x['logP'] = self.model.log_prior(x)
        log_q = self.log_proposal(x)
        log_w = x['logP'] - log_q
        log_w -= np.nanmax(log_w)
        return log_w

    def populate(self, N=None):
        """
        Populate the pool by drawing from the proposal distribution and
        using rejection sampling.
        """
        if N is None:
            N = self.poolsize
        x = self.draw_proposal()
        log_w = self.compute_weights(x)
        log_u = np.log(np.random.rand(N))
        indices = np.where((log_w - log_u) >= 0)[0]
        self.samples = x[indices]
        self.indices = np.random.permutation(self.samples.shape[0]).tolist()
        self.population_acceptance = self.samples.size / self.poolsize
        if self.pool is not None:
            self.evaluate_likelihoods()
        self.populated = True
        self._checked_population = False

    def draw(self, old_sample):
        """
        Propose a new sample. Draws from the pool if it is populated, else
        it populates the pool.

        Parameters
        ----------
        old_sample : structured_array
            Old sample, this is not used in the proposal method
        """
        if not self.populated:
            st = datetime.datetime.now()
            self.populate()
            self.population_time += (datetime.datetime.now() - st)
        index = self.indices.pop()
        new_sample = self.samples[index]
        if not self.indices:
            self.populated = False
        return new_sample



# -*- coding: utf-8 -*-
"""
Base object for all proposal classes.
"""
import datetime
import logging

logger = logging.getLogger(__name__)


def _initialize_global_variables(model):
    """
    Store a global copy of the model for multiprocessing.
    """
    global _model
    _model = model


def _log_likelihood_wrapper(x):
    """
    Wrapper for the log likelihood
    """
    return _model.evaluate_log_likelihood(x)


class Proposal:
    """
    Base proposal object

    Parameters
    ----------
    model: obj
        User-defined model
    """
    def __init__(self, model, n_pool=None):
        self.model = model
        self.populated = True
        self._initialised = False
        self.training_count = 0
        self.population_acceptance = None
        self.population_time = datetime.timedelta()
        self.logl_eval_time = datetime.timedelta()
        self.r = None
        self.n_pool = n_pool
        self.samples = []
        self.indices = []
        self.pool = None
        self._checked_population = True

    @property
    def initialised(self):
        """Boolean that indicates if the proposal is initialised or not."""
        return self._initialised

    @initialised.setter
    def initialised(self, boolean):
        """Setter for initialised

        If value is set to true, the proposal method is tested with `test_draw`
        """
        if boolean:
            self._initialised = boolean
            # TODO: make this useable
            # self.test_draw()
        else:
            self._initialised = boolean

    def initialise(self):
        """
        Initialise the proposal
        """
        self.initialised = True

    def configure_pool(self):
        """
        Configure the multiprocessing pool
        """
        if self.pool is None and self.n_pool is not None:
            if hasattr(self, 'check_acceptance') and not self.check_acceptance:
                self.check_acceptance = True
            logger.info(
                f'Starting multiprocessing pool with {self.n_pool} processes')
            import multiprocessing
            self.pool = multiprocessing.Pool(
                processes=self.n_pool,
                initializer=_initialize_global_variables,
                initargs=(self.model,)
            )
        else:
            logger.info('n_pool is none, no multiprocessing pool')

    def close_pool(self, code=None):
        """
        Close the the multiprocessing pool
        """
        if getattr(self, "pool", None) is not None:
            logger.info("Starting to close worker pool.")
            if code == 2:
                self.pool.terminate()
            else:
                self.pool.close()
            self.pool.join()
            self.pool = None
            logger.info("Finished closing worker pool.")

    def evaluate_likelihoods(self):
        """
        Evaluate the likelihoods for the pool of live points.

        If the multiprocessing pool has been started, the samples will be map
        using `pool.map`.
        """
        st = datetime.datetime.now()
        if self.pool is None:
            for s in self.samples:
                s['logL'] = self.model.evaluate_log_likelihood(s)
        else:
            self.samples['logL'] = self.pool.map(_log_likelihood_wrapper,
                                                 self.samples)
            self.model.likelihood_evaluations += self.samples.size

        self.logl_eval_time += (datetime.datetime.now() - st)

    def draw(self, old_param):
        """
        New a new point given the old point
        """
        raise NotImplementedError

    def test_draw(self):
        """
        Test the draw method to ensure it returns a sample in the correct
        format and the the log prior is computed.
        """
        logger.debug(f'Testing {self.__class__.__name__} draw method')

        test_point = self.model.new_point()
        new_point = self.draw(test_point)

        if new_point['logP'] != self.model.log_prior(new_point):
            raise RuntimeError('Log prior of new point is incorrect!')

        logger.debug(f'{self.__class__.__name__} passed draw test')

    def train(self, x, **kwargs):
        """
        Train the proposal method

        Parameters
        ----------
        x: array_like
            Array of live points to use for training
        kwargs:
            Any of keyword arguments
        """
        logger.info('This proposal method cannot be trained')

    def resume(self, model):
        """
        Resume the proposal with the model
        """
        self.model = model

    def __getstate__(self):
        state = self.__dict__.copy()
        state['pool'] = None
        del state['model']
        return state
